import heapq
from collections import defaultdict, Counter
import dataclasses
from typing import Any, Union, List, Tuple, Dict, Set, Tuple, Sequence
from itertools import product

from flippy.core import ProgramState, ReturnState, SampleState, ObserveState
from flippy.interpreter import CPSInterpreter
from flippy.distributions import Categorical
from flippy.tools import logsumexp, softmax_dict
from flippy.map import MapEnter
from flippy.callentryexit import EnterCallState, ExitCallState
from flippy.types import Element
from flippy.inference.inference import InferenceAlgorithm

@dataclasses.dataclass(order=True)
class PrioritizedItem:
    priority: int
    item: Any=dataclasses.field(compare=False)

@dataclasses.dataclass(frozen=True)
class ProgramStateRecord:
    kind: type
    name: str

@dataclasses.dataclass
class EnumerationStats:
    states_visited: list[ProgramStateRecord] = dataclasses.field(default_factory=list)

    def site_counts(self):
        return Counter(self.states_visited)

class SimpleEnumeration(InferenceAlgorithm[Element]):
    """
    Enumerates all possible executions of a program with discrete random variables
    as a tree, collecting the scores (log-probabilities) of return states.
    - `function`: The function to be executed
    - `max_executions`: The maximum number of executions to consider (default is infinite)
    """
    is_cachable = True  # Inference results can always be cached

    def __init__(self, function, max_executions=float('inf')):
        self.function = function
        self.max_executions = max_executions
        self._stats = None

    def enumerate_tree(
        self,
        ps: ProgramState,
        max_executions: int,
    ) -> Dict[Any, float]:
        frontier: List[PrioritizedItem] = []
        return_scores = {}
        executions = 0
        heapq.heappush(frontier, PrioritizedItem(0, ps))
        while len(frontier) > 0:
            if executions >= max_executions:
                break
            item = heapq.heappop(frontier)
            cum_weight = -item.priority
            ps = item.item
            if isinstance(ps, SampleState):
                for value in ps.distribution.support:
                    weight = ps.distribution.log_probability(value)
                    if weight > float('-inf'):
                        new_ps = ps.step(value)
                        heapq.heappush(frontier, PrioritizedItem(-(cum_weight + weight), new_ps))
            elif isinstance(ps, ObserveState):
                value = ps.value
                weight = ps.distribution.log_probability(value)
                if weight > float('-inf'):
                    new_ps = ps.step()
                    heapq.heappush(frontier, PrioritizedItem(-(cum_weight + weight), new_ps))
            elif isinstance(ps, ReturnState):
                return_scores[ps.value] = logsumexp(
                    return_scores.get(ps.value, float('-inf')),
                    cum_weight
                )
                executions += 1
            elif isinstance(ps, (EnterCallState, ExitCallState, MapEnter)):
                new_ps = ps.step()
                heapq.heappush(frontier, PrioritizedItem(-cum_weight, new_ps))
            else:
                raise ValueError(f"Unrecognized program state message, {ps}")
            if self._stats is not None:
                self._stats.states_visited.append(ProgramStateRecord(ps.__class__, ps.name))
        return return_scores

    def run(self, *args, **kws) -> Categorical[Element]:
        init_ps = CPSInterpreter().initial_program_state(self.function)
        ps = init_ps.step(*args, **kws)
        return_scores = self.enumerate_tree(ps, self.max_executions)
        if len(return_scores) == 0:
            raise ValueError("No return states encountered during enumeration")
        return Categorical.from_dict(softmax_dict(return_scores))

    def _run_with_stats(self, *args, **kws) -> Tuple[Categorical, EnumerationStats]:
        self._stats = EnumerationStats()
        result = self.run(*args, **kws)
        self._stats, stats = None, self._stats
        return result, stats



import linecache
from joblib import Parallel, delayed

from flippy.inference import Enumeration, LikelihoodWeighting
from flippy.interpreter import CPSInterpreter
from flippy.transforms import CPSFunction
from flippy.distributions.builtin_dists import Uniform, Bernoulli, Categorical
from flippy import flip, draw_from, condition, keep_deterministic, hashabledict
from flippy.tools import isclose

def g(p):
    return flip(p) + flip(p) + flip(p) + flip(p) + flip(p)

def f(p=.5):
    Bernoulli(.9).observe(True) # this will only affect the marginal likelihood
    i = draw_from(range(5))
    j = g(p)
    condition(.1 if i == j else .5)
    return i+j

def h():
    p = Uniform().sample()
    return f(p)

def test_parallel_Enumeration():
    cpu1_dist = Enumeration(f, _cpus=1).run()
    cpu2_dist = Enumeration(f, _cpus=2).run()
    assert cpu1_dist.isclose(cpu2_dist)
    assert isclose(cpu1_dist.marginal_likelihood, cpu2_dist.marginal_likelihood, atol=1e-5)

def test_parallel_LikelihoodWeighting():
    seed = 1391
    cpu1_dist = LikelihoodWeighting(h, samples=5000, _cpus=1, seed=seed).run()
    cpu1_dist = cpu1_dist.marginalize(lambda x: x % 2)
    cpu2_dist = LikelihoodWeighting(h, samples=5000, _cpus=2, seed=seed).run()
    cpu2_dist = cpu2_dist.marginalize(lambda x: x % 2)
    assert cpu1_dist.isclose(cpu2_dist, atol=5e-2), (cpu1_dist, cpu2_dist)
    assert not cpu1_dist.isclose(cpu2_dist, atol=1e-5)

def test_Enumeration__run_partition_enumerates_properly():
    sample_count = [0]
    variable_list = []

    @keep_deterministic
    def inc_sample_count(**variables):
        variable_list.append(hashabledict(variables))
        sample_count[0] += 1

    # special hook for debugging
    inc_sample_count._emit_call_entryexit = False

    def f(p):
        a = Bernoulli(p).sample()
        inc_sample_count(a=a)
        b = Bernoulli(p).sample()
        inc_sample_count(a=a, b=b)
        return a + b

    def run_partitions(f, partitions):
        scs = []
        vls = []
        for idx in range(partitions):
            sample_count[0] = 0
            variable_list[:] = []
            Enumeration(f)._run_partition(.7, _partition_idx=idx, _partitions=partitions, _linecache=linecache.cache)
            scs.append(sample_count[0])
            vls.append(list(variable_list))
        return scs, vls

    scs, vls = run_partitions(f, 2)
    assert scs == [2 + 2] * 2
    prefix = [dict(a=1), dict(a=0)]
    assert vls == [
        prefix + [dict(a=1, b=1), dict(a=1, b=0)],
        prefix + [dict(a=0, b=1), dict(a=0, b=0)],
    ]
    # Checking for completeness, though we need to avoid double-counting the shared prefix
    assert sum(scs) - len(prefix) == len(set([s for partition in vls for s in partition]))

    def f3(p):
        a = Bernoulli(p).sample()
        inc_sample_count(a=a)
        b = Bernoulli(p).sample()
        inc_sample_count(a=a, b=b)
        c = Bernoulli(p).sample()
        inc_sample_count(a=a, b=b, c=c)
        return a + b + c

    # Checking case where frontier exceeds partition count after partitioning
    scs, vls = run_partitions(f3, 2)
    assert scs == [2 + 2 + 4] * 2
    prefix = [dict(a=1), dict(a=0)]
    assert vls == [
        prefix + [dict(a=1, b=1), dict(a=1, b=0)] + [dict(a=1, b=b, c=c) for b in [1, 0] for c in [1, 0]],
        prefix + [dict(a=0, b=1), dict(a=0, b=0)] + [dict(a=0, b=b, c=c) for b in [1, 0] for c in [1, 0]],
    ]
    assert sum(scs) - len(prefix) == len(set([s for partition in vls for s in partition]))

    # Checking case where number of partitions is different than program's branching factor
    scs, vls = run_partitions(f3, 3)
    assert scs == [6, 10, 6]
    prefix = [dict(a=1), dict(a=0), dict(a=1, b=1), dict(a=1, b=0)]
    assert vls == [
        prefix + [dict(a=1, b=1, c=c) for c in [1, 0]],
        prefix + [dict(a=0, b=1), dict(a=0, b=0)] + [dict(a=0, b=b, c=c) for b in [1, 0] for c in [1, 0]],
        prefix + [dict(a=1, b=0, c=c) for c in [1, 0]],
    ]
    assert sum(scs) - 2 * len(prefix) == len(set([s for partition in vls for s in partition]))

def test_CPSFunction_can_be_passed_to_joblib_loky_worker():
    i = 100
    def f():
        return 100 + i
    def m(cps_f2: CPSFunction):
        assert cps_f2()() == 200
    cps_f = CPSInterpreter().non_cps_callable_to_cps_callable(f)
    assert cps_f()() == 200
    Parallel(n_jobs=2, backend="loky")(delayed(m)(cps_f) for _ in range(2))



from typing import Sequence
from functools import cached_property

from flippy.distributions.base import Distribution, Element
from flippy.distributions.random import RandomNumberGenerator, default_rng
from flippy.tools import isclose, PackagePlaceholder

try:
    import scipy.stats as scipy_stats
except ImportError:
    scipy_stats = PackagePlaceholder("scipy.stats")
try:
    import numpy as np
except ImportError:
    np = PackagePlaceholder("numpy")
try:
    import sympy as sp
except ImportError:
    sp = PackagePlaceholder("sympy")


class MultivariateNormal(Distribution):
    """Multivariate normal distribution with given means and covariance matrix."""
    def __init__(self, means=(0,), covariance=((1,),)):
        assert len(means) == len(covariance), "Means and covariance must have the same length"
        assert all(len(row) == len(means) for row in covariance), "Covariance must be a square matrix matching the length of means"
        self.means = means
        self.covariance = covariance

    dim = property(lambda self: len(self.means))
    support = cached_property(lambda self: sp.Reals**self.dim)

    def sample(self, rng=default_rng, name=None, initial_value=None) -> float:
        x = scipy_stats.multivariate_normal.rvs(mean=self.means, cov=self.covariance, random_state=rng.np)
        if isinstance(x, (int, float)):
            return (x,)
        return tuple(x)

    def log_probability(self, element):
        if any(isinstance(x, sp.Basic) for x in [element, self.means, self.covariance]):
            raise NotImplementedError("Symbolic computation not supported for MultivariateNormal")
        return scipy_stats.multivariate_normal.logpdf(element, mean=self.means, cov=self.covariance)

class InverseWishart(Distribution):
    """Inverse Wishart distribution with given degrees of freedom and scale matrix."""
    def __init__(self, df=1, scale_matrix=((1,),)):
        scale_matrix = np.array(scale_matrix)
        assert scale_matrix.ndim == 2, "Scale matrix must be a 2D array"
        assert scale_matrix.shape[0] == scale_matrix.shape[1], \
            "Scale matrix must be square"
        assert df >= scale_matrix.shape[0], \
            "Degrees of freedom must be at least the dimension of the scale matrix"
        self.df = df
        self.scale_matrix = scale_matrix

    dim = property(lambda self: len(self.scale_matrix))
    support = cached_property(lambda self: (sp.Reals**self.dim)**self.dim)

    def sample(self, rng=default_rng, name=None, initial_value=None) -> float:
        x = scipy_stats.invwishart.rvs(df=self.df, scale=self.scale_matrix, random_state=rng.np)
        if isinstance(x, (int, float)):
            return x
        return tuple(tuple(xi) for xi in x)

    def log_probability(self, element):
        if any(isinstance(x, sp.Basic) for x in [element, self.df, self.scale_matrix]):
            raise NotImplementedError("Symbolic computation not supported for MultivariateNormal")
        return scipy_stats.invwishart.logpdf(element, df=self.df, scale=self.scale_matrix)

    def expected_value(self, func = None):
        assert func is None, "Arbitrary expected value function not implemented for InverseWishart"
        if self.df <= self.dim + 1:
            raise ValueError("Expected value is not defined for df <= dim + 1")
        return self.scale_matrix / (self.df - self.dim - 1)


class NormalNormal(Distribution):
    """
    Compound distribution for a normal distribution
    with unknown mean and known variance. The prior on the mean is
    also a normal distribution, which is its own conjugate distribution.
    """
    def __init__(self, *, prior_mean=0, prior_sd=1, sd=1):
        self.prior_mean = np.array(prior_mean)
        self.prior_sd = np.array(prior_sd)
        self.sd = np.array(sd)
    marginal_sd = property(lambda self: (self.prior_sd**2 + self.sd**2)**.5)

    def sample(
        self,
        rng : RandomNumberGenerator = default_rng,
        name=None,
        initial_value=None
    ) -> Sequence[Element]:
        #first sample mu and then sample x
        #see scipy norm documentation for information on loc and scale
        mean = scipy_stats.norm.rvs(loc=self.prior_mean, scale=self.prior_sd, random_state=rng.np)
        x = scipy_stats.norm.rvs(loc=mean, scale=self.sd, random_state=rng.np)
        return x

    def log_probability(self, element : Sequence[Element]) -> float:
        return scipy_stats.norm.logpdf(element, loc=self.prior_mean, scale=self.marginal_sd)

    def update(self, data : Element | Sequence[Element]) -> "NormalNormal":
        #gets posterior predictive term
        if isinstance(data, (float, int)):
            data = [data]  # Ensure data is a sequence
        total = sum(data)
        n_datapoints = len(data)
        new_prior_var = 1/(1/self.prior_sd**2 + n_datapoints/self.sd**2)
        new_prior_sd = new_prior_var**.5
        new_prior_mean = (self.prior_mean/self.prior_sd**2 + total/self.sd**2) * new_prior_var
        return NormalNormal(prior_mean=new_prior_mean, prior_sd=new_prior_sd, sd=self.sd)

class MultivariateNormalNormal(Distribution):
    """
    Compound distribution for a multivariate normal distribution
    with unknown mean and known covariance. The prior on the mean is
    also a multivariate normal distribution, which is its own conjugate distribution.
    """
    def __init__(
        self,
        *,
        prior_means : Sequence[float] = (0,),
        prior_cov: Sequence[Sequence[float]] = ((1,),),
        cov: Sequence[Sequence[float]] = ((1,),),
    ):
        self.prior_means = np.array(prior_means)
        self.prior_cov = np.array(prior_cov)
        self.cov = np.array(cov)
        assert len(np.shape(self.prior_cov)) == 2 #make sure cov is 2d
        assert len(np.shape(self.cov)) == 2 #make sure cov is 2d
        assert np.shape(self.prior_means)[0] == np.shape(self.prior_cov)[0] == \
            np.shape(self.prior_cov)[1] == np.shape(self.cov)[0] == \
            np.shape(self.cov)[1], "Means and covariances must match in dimensions"

    def sample(
        self,
        rng : RandomNumberGenerator = default_rng,
        name=None,
        initial_value=None
    ) -> Sequence[Element]:
        mean = scipy_stats.multivariate_normal.rvs(self.prior_means, self.prior_cov, random_state=rng.np)
        x = scipy_stats.multivariate_normal.rvs(mean, self.cov, random_state=rng.np)
        return x

    def log_probabilities(self, element : Sequence[Element]) -> float:
        # Reference: https://www.cs.ubc.ca/~murphyk/Papers/bayesGauss.pdf
        element = np.array(element)
        assert np.ndim(element) in (1, 2), "Element must be a 1D or 2D array"
        if np.ndim(element) == 1:
            element = element.reshape(1, -1)
        assert np.shape(element)[1] == len(self.prior_means), \
            f"Each element must have shape {len(self.prior_means)}, got {np.shape(element)}"
        marginal_cov = self.prior_cov + self.cov
        return scipy_stats.multivariate_normal.logpdf(element, mean=self.prior_means, cov=marginal_cov)

    def log_probability(self, element : Sequence[Element]) -> float:
        return self.log_probabilities(element)

    def update(self, data : Sequence[Element]) -> "MultivariateNormalNormal":
        raise NotImplementedError
        # if isinstance(data[0], (float, int)):
        #     total = data
        #     n_datapoints = 1
        # elif isinstance(data[0][0], (float, int)):
        #     total = np.sum(data,axis=0)
        #     n_datapoints = len(data)
        # else:
        #     raise ValueError(f"Invalid data shape {data}")
        # prior_cov_inverted = np.linalg.inv(self.prior_cov)
        # cov_inverted = np.linalg.inv(self.cov)

        # new_prior_cov = np.linalg.inv(prior_cov_inverted + n_datapoints * cov_inverted)
        # new_prior_means = new_prior_cov @ (cov_inverted @ total + prior_cov_inverted @ self.prior_means)
        # return MultivariateNormalNormal(prior_means=new_prior_means, prior_cov=new_prior_cov, cov=self.cov, size=self.size)



import math
import linecache
import joblib
import pytest

from flippy import flip, mem, infer, draw_from, factor, condition, \
    Bernoulli, Categorical, Uniform, keep_deterministic,\
    uniform, recursive_map, recursive_filter, recursive_reduce, \
    map_observe, cps_transform_safe_decorator
from flippy.distributions.builtin_dists import Bernoulli, Uniform, Binomial
from flippy.interpreter import CPSInterpreter
from flippy.inference import SimpleEnumeration, LikelihoodWeighting, Enumeration, \
    MaximumMarginalAPosteriori, MetropolisHastings, SamplePrior
from flippy.core import ReturnState
from flippy.tools import isclose


def algebra():
    def flip():
        return Bernoulli(0.5).sample()
    def NUM():
        return Categorical(range(5)).sample()
    def OP():
        if flip():
            return '+'
        else:
            return '*'
    def EQ():
        if flip():
            return NUM()
        else:
            return (NUM(), OP(), EQ())
    return EQ()

def test_algebra():
    algebra2 = infer(method="SimpleEnumeration", max_executions=55)(algebra)

    results = algebra2()

    num_ct = 0
    expr_ct = 0
    for e in results.support:
        if isinstance(e, int):
            num_ct += 1
        else:
            assert (
                isinstance(e, tuple) and
                len(e) == 3 and
                isinstance(e[0], int) and
                e[1] in ('*', '+') and
                isinstance(e[2], int)
            )
            expr_ct += 1
    assert num_ct == 5
    assert expr_ct == 5 * 2 * 5

def world_prior():
    return Categorical(range(4)).sample()

def utterance_prior():
    return Categorical([
        'some of the people are nice',
        'all of the people are nice',
        'none of the people are nice',
    ]).sample()

def meaning(utt, world):
    return (
        world > 0 if utt == 'some of the people are nice' else
        world == 3 if utt == 'all of the people are nice' else
        world == 0 if utt == 'none of the people are nice' else
        True
    )

@infer
def literal_listener(utterance):
    world = world_prior()
    m = meaning(utterance, world)
    Bernoulli(1.0).observe(m)
    return world

@infer
def speaker(world) -> str:
    utterance = utterance_prior()
    L = literal_listener(utterance)
    L.observe(world)
    return utterance

@infer
def listener(utterance):
    world = world_prior()
    S = speaker(world)
    S.observe(utterance)
    return world

def test_scalar_implicature():
    assert listener('some of the people are nice').isclose(Categorical(
        [1, 2, 3],
        probabilities=[4/9, 4/9, 1/9],
    ))

def test_planning_as_inference():
    import math
    class MDP:
        '''
        This simple task has an always-increasing state, until a terminal
        limit is reached. Rewards are directly based on action indices
        selected.
        '''
        def __init__(self, limit=2):
            self.limit = limit
        def initial_state(self):
            return 0
        def actions(self):
            return [0, 1]
        def is_terminal(self, s):
            return s == self.limit
        def next_state(self, s, a):
            return s + 1
        def reward(self, s, a, ns):
            return -a
    @infer
    def fn(mdp):
        s = mdp.initial_state()
        actions = ()
        while not mdp.is_terminal(s):
            a = Categorical(mdp.actions()).sample()
            actions += (a,)
            ns = mdp.next_state(s, a)
            r = mdp.reward(s, a, ns)
            assert r <= 0
            Bernoulli(math.exp(r)).observe(1)
            s = ns
        return actions
    # Enumerate all action sequences
    s = [(a, b) for a in range(2) for b in range(2)]
    # Total reward for a sequence is exp(-sum(actions))
    expected = Categorical(s, weights=[math.exp(-sum(ev)) for ev in s])
    assert fn(MDP()).isclose(expected)

def test_infer():
    def f0():
        return Bernoulli(.4).sample()
    f0 = infer(f0)
    @infer
    def f1():
        return Bernoulli(.4).sample()
    @infer()
    def f2():
        return Bernoulli(.4).sample()
    @infer(cache_size=10)
    def f3():
        return Bernoulli(.4).sample()
    assert f1().isclose(f2())
    assert f1().isclose(f0())
    assert f1().isclose(f3())


def test_builtins():
    @infer
    def model():
        return abs(Categorical([-1, 0, 1]).sample())
    assert model().isclose(Categorical(
        [0, 1],
        probabilities=[1/3, 2/3],
    ))

    @infer
    def model():
        mydict = {}
        return tuple(mydict.items())
    assert model().isclose(Categorical([(),]))

def test_cps_map():
    assert recursive_map(lambda x: x ** 2, []) == []
    assert recursive_map(lambda x: x ** 2, list(range(5))) == [0, 1, 4, 9, 16]

def test_cps_filter():
    assert recursive_filter(lambda x: x % 2 == 0, []) == []
    assert recursive_filter(lambda x: x % 2 == 0, list(range(5))) == [0, 2, 4]

def test_recursive_reduce():
    sumfn = lambda acc, el: acc + el

    assert recursive_reduce(sumfn, [], 0) == 0
    assert recursive_reduce(sumfn, [1], 0) == 1
    assert recursive_reduce(sumfn, [1, 2], 0) == 3
    assert recursive_reduce(sumfn, [1, 2, 3], 0) == 6

    assert recursive_reduce(sumfn, [[3, 4], [5]], []) == [3, 4, 5]

def test_stochastic_memoization():
    def stochastic_mem_func():
        def g(i):
            return flip()
        g = mem(g)
        return (g(1), g(1), g(2), g(2), flip())

    def no_stochastic_mem_func():
        def g(i):
            return flip()
        return (g(1), g(1), g(2), g(2), flip())

    assert len(SimpleEnumeration(stochastic_mem_func).run().support) == 2**3
    assert len(SimpleEnumeration(no_stochastic_mem_func).run().support) == 2**5


def test_mem_basic():
    def with_mem():
        def f(i):
            return Uniform(0, 1).sample()
        f = mem(f)
        x = f(0)
        return (x + x, f(0) + f(0))

    u = .456
    ps = CPSInterpreter().initial_program_state(with_mem)
    ps = ps.step().step(u)
    assert isinstance(ps, ReturnState)
    assert ps.value == (u*2, u*2)

    def without_mem():
        def f(i):
            return Uniform(0, 1).sample()
        x = f(0)
        return (x + x, f(0) + f(0))

    ps = CPSInterpreter().initial_program_state(without_mem)
    ps = ps.step().step(u).step(0).step(0)
    assert isinstance(ps, ReturnState)
    assert ps.value == (u*2, 0)


def test_draw_from():
    def f():
        return draw_from(3)
    assert SimpleEnumeration(f).run().isclose(Categorical.from_dict({0: 1/3, 1: 1/3, 2: 1/3}))

def test_factor_statement():
    def f():
        x = flip(.4)
        y = flip(.7)
        factor(math.log(.2) if x == y else math.log(.8))
        return (x, y)

    exp = {
        (0, 0): .2*.6*.3,
        (0, 1): .8*.6*.7,
        (1, 0): .8*.4*.3,
        (1, 1): .2*.4*.7,
    }
    exp = Categorical.from_dict({k: v/sum(exp.values()) for k, v in exp.items()})
    assert SimpleEnumeration(f).run().isclose(exp)

    def f_with_positive():
        x = flip()
        factor(1.23 if x else -1)
        return x
    exp = {
        True: math.exp(1.23),
        False: math.exp(-1),
    }
    exp = Categorical.from_dict({k: v/sum(exp.values()) for k, v in exp.items()})
    assert SimpleEnumeration(f_with_positive).run().isclose(exp)

def test_condition_statement():
    def f():
        u = uniform()
        condition(.25 < u < .75)
        return u
    samples = LikelihoodWeighting(f, samples=1000).run().support
    assert all(.25 < s < .75 for s in samples)

def test_map_observe():
    @infer
    def f1():
        p = .1 if Bernoulli(.5).sample() else .9
        map_observe(Bernoulli(p), [1, 1, 0, 1])
        return p

    @infer
    def f2():
        p = .1 if Bernoulli(.5).sample() else .9
        [Bernoulli(p).observe(i) for i in [1, 1, 0, 1]]
        return p
    assert f2().isclose(f1())

def test_nested_decorators():
    # module-scope decorator only
    @infer
    def model():
        def f(p):
            return flip(p)
        return f(.2)

    # module-scoped decorator + function-scoped decorator with different return type
    @infer
    def model2a():
        @infer
        def f(p):
            return flip(p)
        return f(.2).sample()
    assert model().isclose(model2a())

    # module-scoped decorator (called) + function-scoped decorator with different return type (called)
    @infer()
    def model2b():
        @infer()
        def f(p):
            return flip(p)
        return f(.2).sample()
    assert model().isclose(model2b())

    # module-scoped decorator + function-scoped decorator with same return type
    @infer
    def model3():
        @mem
        def f(p):
            return flip(p)
        return f(.2)
    assert model().isclose(model3())

def test_sibling_decorators():
    @infer
    def model():
        def f(p):
            return flip(p)
        return f(.2)

    def outer_f(p):
        return flip(p)
    @infer
    def model4():
        return outer_f(.2)
    assert model().isclose(model4())

    @infer
    def outer_f_infer(p):
        return flip(p)
    @infer
    def model5():
        return outer_f_infer(.2).sample()
    assert model().isclose(model5())

    @mem
    def outer_f_mem(p):
        return flip(p)
    @infer
    def model6():
        return outer_f_mem(.2)
    assert model().isclose(model6())

def test_chained_decorators():
    @cps_transform_safe_decorator
    def dec1(fn):
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs) + 'a'
        return wrapper

    @cps_transform_safe_decorator
    def dec2(fn):
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs) + 'b'
        return wrapper

    @infer
    def model_12():
        @dec2
        @dec1
        def f(p):
            x = '1' if flip(p) else '0'
            return x + "_"
        return f(0.4)

    @infer
    def model_21():
        @dec1
        @dec2
        def f(p):
            x = '1' if flip(p) else '0'
            return x + "_"
        return f(0.4)

    assert model_12().isclose(Categorical.from_dict({'0_ab': 0.6, '1_ab': 0.4}))
    assert model_21().isclose(Categorical.from_dict({'0_ba': 0.6, '1_ba': 0.4}))

def test_keep_deterministic__wrapping_function():
    @keep_deterministic
    def function(i):
        return i + flip()
    assert function(1) in (1, 2)
    assert function(2) in (2, 3)

def test_keep_deterministic__wrapping_methods():
    def tests(MyClass):
        c = MyClass()
        assert c.normal_method(1) in (100, 101)
        assert MyClass.normal_method(c, 1) in (100, 101)

        c2 = MyClass(200)
        assert c2.normal_method(1) in (200, 201)
        assert MyClass.normal_method(c2, 1) in (200, 201)

        assert c.class_method(1) in (100, 101)
        assert c2.class_method(1) in (100, 101)
        assert MyClass.class_method(1) in (100, 101)

        assert c.static_method(1) in (100, 101)
        assert c2.static_method(1) in (100, 101)
        assert MyClass.static_method(1) in (100, 101)

    class MyClass:
        x = 100
        def __init__(self, x=None):
            if x is not None:
                self.x = x

        @keep_deterministic
        def normal_method(self, i):
            return flip()*i + self.x

        @keep_deterministic
        @classmethod
        def class_method(cls, i):
            return flip()*i + cls.x

        @keep_deterministic
        @staticmethod
        def static_method(i):
            return flip()*i + MyClass.x

    tests(MyClass)

    # now test calling method from within model
    @infer
    def model():
        tests(MyClass)
    model()


def test_infer_wrapping_instance_method():
    class MyClass:
        @infer
        def stochastic_method(self):
            return flip(.65)

    @infer
    def stochastic_function():
        return flip(.65)

    c = MyClass()

    assert c.stochastic_method().isclose(stochastic_function())
    assert MyClass.stochastic_method(c).isclose(stochastic_function())

    # now test calling method from within model
    @infer
    def model_1():
        return c.stochastic_method().sample()
    assert model_1().isclose(stochastic_function())

    @infer
    def model_2():
        return MyClass.stochastic_method(c).sample()
    assert model_2().isclose(stochastic_function())

def test_keep_deterministic_wrapping_instance_method():
    class MyClass:
        @keep_deterministic
        def deterministic_method(self):
            return 1

    c = MyClass()
    assert c.deterministic_method() == 1
    assert MyClass.deterministic_method(c) == 1

    # now test calling method from within model
    @infer
    def model():
        c = MyClass()
        x = c.deterministic_method()
        y = MyClass.deterministic_method(c)
        return x + y

    assert model().isclose(Categorical.from_dict({2: 1}))

def test_infer__called_by_function_in_joblib_Parallel_loky():
    @infer(method=Enumeration, cache_size=0)
    def model():
        return 123

    def f(_linecache):
        linecache.cache.update(_linecache)
        return model()

    res = joblib.Parallel(n_jobs=1, backend="loky")(joblib.delayed(f)(linecache.cache) for _ in range(2))
    assert all([r.isclose(Categorical([123])) for r in res])

def test_infer__called_by_function_in_joblib_Parallel_loky_with_caching():
    @infer(method=SimpleEnumeration, cache_size=5)
    def model(p):
        return flip(p) + flip(p)

    model(.5)
    model(.7)

    def f(p, linecache_cache):
        linecache.cache.update(linecache_cache)
        assert len(model.cache) == 2
        model(p)
        if p not in (.5, .7):
            assert len(model.cache) == 3
        else:
            assert len(model.cache) == 2
        return model(p)

    joblib.Parallel(n_jobs=2)(joblib.delayed(f)(p, linecache.cache) for p in [.5, .6, .7, .8])
    assert len(model.cache) == 2

def test_infer__multi_cpu_method_decorator():
    @infer(method=Enumeration, _cpus=2)
    def f_dec() -> bool:
        return flip(.67)
    assert f_dec().isclose(Bernoulli(.67))

def test_infer__multi_cpu_method_nondecorator():
    def f_nodec():
        return flip(.67)
    f_nodec = infer(f_nodec, method=Enumeration, _cpus=2)
    assert f_nodec().isclose(Bernoulli(.67))

# Combinations of situations in which we want to to test infer on methods:
# - is infer being used as a decorator?
# - is infer applied to a normal method / classmethod / staticmethod?
# - is the inferred method accessed from instance or from the class?
# - is infer using a multi-cpu inference method?
# - is the inferred method being used in some other parallelized code?
# - is infer applied inside class or outside class scope? (see notes below)

def _test_infer__on_class_defined_method_serial(MyClass):
    @infer
    def stochastic_function():
        return flip(.65)

    c = MyClass()
    assert c.normal_method().isclose(stochastic_function())
    assert c.cls_method().isclose(stochastic_function())
    assert c.static_method().isclose(stochastic_function())

    assert MyClass.normal_method(c).isclose(stochastic_function())
    assert MyClass.cls_method().isclose(stochastic_function())
    assert MyClass.static_method().isclose(stochastic_function())

def _test_infer__on_class_defined_method_parallel(MyClass):
    def parallelized_func(linecache_):
        @infer
        def stochastic_function():
            return flip(.65)
        linecache.cache.update(linecache_)
        c = MyClass()
        assert c.normal_method().isclose(stochastic_function())
        assert c.cls_method().isclose(stochastic_function())
        assert c.static_method().isclose(stochastic_function())

        assert MyClass.normal_method(c).isclose(stochastic_function())
        assert MyClass.cls_method().isclose(stochastic_function())
        assert MyClass.static_method().isclose(stochastic_function())
    joblib.Parallel(n_jobs=2, backend='loky')(
        joblib.delayed(parallelized_func)(linecache.cache) for _ in range(2)
    )

def _test_infer__on_class_defined_method_nested(MyClass):
    @infer(method=SimpleEnumeration)
    def model():
        c = MyClass()
        assert c.normal_method().sample() in (0, 1)
        assert c.cls_method().sample() in (0, 1)
        assert c.static_method().sample() in (0, 1)
        assert MyClass.normal_method(c).sample() in (0, 1)
        assert MyClass.cls_method().sample() in (0, 1)
        assert MyClass.static_method().sample() in (0, 1)
        return 100
    assert model().isclose(Categorical([100]))

def _test_infer__on_class_defined_method_nested_multicpu_method(MyClass):
    @infer(method=Enumeration, _cpus=2)
    def model():
        c = MyClass()
        assert c.normal_method().sample() in (0, 1)
        assert c.cls_method().sample() in (0, 1)
        assert c.static_method().sample() in (0, 1)
        assert MyClass.normal_method(c).sample() in (0, 1)
        assert MyClass.cls_method().sample() in (0, 1)
        assert MyClass.static_method().sample() in (0, 1)
        return 100
    assert model().isclose(Categorical([100]))

def test_infer__decorating_methods__inside_class():
    class MyClass:
        @infer
        def normal_method(self, p=.65):
            return flip(p)

        @infer
        @classmethod
        def cls_method(cls, p=.65):
            return flip(p)

        @infer
        @staticmethod
        def static_method(p=.65):
            return flip(p)
    _test_infer__on_class_defined_method_serial(MyClass)
    _test_infer__on_class_defined_method_parallel(MyClass)
    _test_infer__on_class_defined_method_nested(MyClass)
    _test_infer__on_class_defined_method_nested_multicpu_method(MyClass)

def test_infer__not_decorating_methods__inside_class():
    class MyClass:
        def normal_method(self, p=.65):
            return flip(p)
        normal_method = infer(normal_method)

        @classmethod
        def cls_method(cls, p=.65):
            return flip(p)
        cls_method = infer(cls_method)

        @staticmethod
        def static_method(p=.65):
            return flip(p)
        static_method = infer(static_method)

    _test_infer__on_class_defined_method_serial(MyClass)
    _test_infer__on_class_defined_method_parallel(MyClass)
    _test_infer__on_class_defined_method_nested(MyClass)
    _test_infer__on_class_defined_method_nested_multicpu_method(MyClass)

def test_infer__decorating_methods__inside_class__multicpu_inference_alg():
    class MyClass:
        @infer(method=Enumeration, _cpus=2)
        def normal_method(self, p=.65):
            return flip(p)

        @infer(method=Enumeration, _cpus=2)
        @classmethod
        def cls_method(cls, p=.65):
            return flip(p)

        @infer(method=Enumeration, _cpus=2)
        @staticmethod
        def static_method(p=.65):
            return flip(p)
    _test_infer__on_class_defined_method_serial(MyClass)
    _test_infer__on_class_defined_method_parallel(MyClass)
    _test_infer__on_class_defined_method_nested(MyClass)
    _test_infer__on_class_defined_method_nested_multicpu_method(MyClass)

def test_infer__applied_to_methods__outside_class_invalid():
    class MyClass:
        def normal_method(self, p=.65):
            return flip(p)

        @classmethod
        def cls_method(cls, p=.65):
            return flip(p)

        @staticmethod
        def static_method(p=.65):
            return flip(p)

    with pytest.raises(ValueError) as e:
        MyClass.cls_method = infer(MyClass.cls_method)
    assert "Cannot wrap a method outside the class namespace" in str(e.value)

    c = MyClass()
    with pytest.raises(ValueError) as e:
        c.normal_method = infer(c.normal_method, method=Enumeration)
    assert "Cannot wrap a method outside the class namespace" in str(e.value)

    # applying infer on a static method outside the class is not supported
    # when its called from an instance
    MyClass.static_method = infer(MyClass.static_method, method=Enumeration)
    c = MyClass()
    with pytest.raises(TypeError) as e:
        c.static_method(.33)
    assert "takes from 0 to 1 positional arguments" in str(e.value)

def test_infer__applied_to_methods__outside_class_valid():
    class MyClass:
        def normal_method(self, p=.65):
            return flip(p)

        @staticmethod
        def static_method(p=.65):
            return flip(p)

    MyClass.normal_method = infer(MyClass.normal_method)
    MyClass.static_method = infer(MyClass.static_method)
    c = MyClass()
    assert c.normal_method(.6).isclose(Bernoulli(.6))
    assert MyClass.normal_method(c, .6).isclose(Bernoulli(.6))
    assert MyClass.static_method(.6).isclose(Bernoulli(.6))

def test_Distribution_fit_interface():
    def param_fit(data):
        p = Uniform(0, 1).fit(initial_value=.123)
        Binomial(len(data), p).observe(sum(data))
        return p

    mmap = MaximumMarginalAPosteriori(param_fit)
    data = (1, 1, 0, 1, 0, 1, 0, 0, 0, 0)*10
    assert param_fit(data) == 0.123
    res = mmap.run(data)
    assert isclose(res.sample(), sum(data)/len(data))

    # These inference algorithms don't support the Distribution.fit interface
    with pytest.raises(AssertionError) as e:
        LikelihoodWeighting(param_fit, samples=100).run(data)
    assert "doesn't support Distribution.fit" in e.value.args[0]

    with pytest.raises(AssertionError) as e:
        MetropolisHastings(param_fit, samples=100).run(data)
    assert "doesn't support Distribution.fit" in e.value.args[0]

    def discrete_fit():
        x = Bernoulli(.5).fit(initial_value=0)
        return x

    with pytest.raises(AssertionError) as e:
        Enumeration(discrete_fit).run()
    assert "doesn't support Distribution.fit" in e.value.args[0]

def test_Distribution_fit_initial_value_interface():
    def param_fit(use_initial_value):
        p = Uniform(0, 1).fit(
            name="p",
            initial_value= .123 if use_initial_value else None
        )
        return p

    mmap = MaximumMarginalAPosteriori(param_fit)
    assignment, _, _ = mmap.assignment_score(args=(True, ), kwargs={}, assignments={})
    assert assignment['p'][0] == .123

    assignment, _, _ = mmap.assignment_score(args=(False, ), kwargs={}, assignments={})
    assert assignment['p'][0] != .123

@infer
def even(x):
    if x == 0:
        return True
    return odd(x - 1).sample()

@infer
def odd(x):
    if x == 0:
        return True
    return even(x - 1).sample()

def test_mutual_recursion():
    @infer
    def nested_even(x):
        if x == 0:
            return True
        return nested_odd(x - 1).sample()

    @infer
    def nested_odd(x):
        if x == 0:
            return True
        return nested_even(x - 1).sample()

    assert even(30).isclose(Categorical([True]))
    assert nested_even(30).isclose(Categorical([True]))

def test_calling_method_from_object_returned_by_function():
    def make_dist():
        p = flip()
        return Bernoulli(p)

    @infer
    def model1():
        return make_dist().sample()

    @infer
    def model2():
        dist = make_dist()
        return dist.sample()

    assert model1() == model2()

def test_is_cachable_property():
    def f():
        return uniform()

    assert Enumeration(f).is_cachable
    assert SimpleEnumeration(f).is_cachable

    assert LikelihoodWeighting(f, samples=10, seed=42).is_cachable
    assert MetropolisHastings(f, samples=10, seed=42).is_cachable
    assert SamplePrior(f, samples=10, seed=42).is_cachable
    assert MaximumMarginalAPosteriori(f, seed=42).is_cachable

    assert not LikelihoodWeighting(f, samples=10, seed=None).is_cachable
    assert not MetropolisHastings(f, samples=10, seed=None).is_cachable
    assert not SamplePrior(f, samples=10, seed=None).is_cachable
    assert not MaximumMarginalAPosteriori(f, seed=None).is_cachable

    not_cachable_f = infer(
        f, method=LikelihoodWeighting, samples=10, seed=None, cache_size=1024
    )
    assert not_cachable_f() != not_cachable_f()

    cachable_f = infer(
        f, method=LikelihoodWeighting, samples=10, seed=42, cache_size=1024
    )
    assert cachable_f() == cachable_f()
    # Stronger test, that value is identical
    assert cachable_f() is cachable_f()

def test_error_for_lambda():
    # Nested lambda works
    @infer
    def fn():
        f = lambda: flip(0.5)
        return f()
    assert fn().isclose(Bernoulli(0.5))

    # Outermost lambda does not
    with pytest.raises(ValueError) as err:
        d = infer(lambda: flip(0.5))()
        print(d)
    assert 'Cannot interpret lambda expressions' in str(err)



"""
This module contains various static transformations and analyses of Python ASTs
that are used for executing Python functions in continuation-passing style (CPS).
"""

import ast
from functools import cached_property
from typing import Callable, TYPE_CHECKING
import textwrap
import copy
import contextlib
import collections
import dataclasses

from flippy.hashable import hashabledict
from flippy.types import CPSCallable

if TYPE_CHECKING:
    from flippy.interpreter import CPSInterpreter

class _ValueWrapper:
    '''
    A simple wrapper class that we use to indicate when a scope is being returned below.
    '''
    def __init__(self, value):
        self.value = value

class Placeholder(ast.NodeTransformer):
    '''
    Use this class to add placeholders in code constructed by string, so that
    placeholders can later be filled by AST instances.
    '''
    prefix = '__PLACEHOLDER_'
    def __init__(self, replacement_map, *, skip_missing=False):
        self.replacement_map = replacement_map
        self.skip_missing = skip_missing

    @classmethod
    def new(cls, tag):
        return f'{cls.prefix}{tag}'
    @classmethod
    def fill(cls, node, replacement_map, **kwargs):
        return cls(replacement_map, **kwargs).visit(node)
    @classmethod
    def fill_from_source(cls, src, *args, **kwargs):
        node = ast.parse(textwrap.dedent(src))
        return cls.fill(node, *args, **kwargs)

    def visit_Name(self, node):
        if node.id.startswith(__class__.prefix):
            tag = node.id[len(__class__.prefix):]
            if tag in self.replacement_map:
                return self.replacement_map[tag]
            elif not self.skip_missing:
                assert tag in self.replacement_map, f'Could not find tag {tag} in replacement map.'
        return self.generic_visit(node)
    def visit_Expr(self, node):
        new_node = self.generic_visit(node)
        if hasattr(new_node, 'value'):
            if isinstance(new_node.value, ast.expr):
                return new_node
            else:
                # This case is when replacement of var was a statement
                return new_node.value
        else:
            # When replacement of var was None, this node should be deleted.
            return None


@dataclasses.dataclass
class Scope:
    # The node this scope is associated with.
    node: ast.AST
    # The parent or outer scope
    parent: 'Scope'
    # This namespace (ns) tracks variable names and the number of times they were written to.
    ns: dict[str, int] = dataclasses.field(default_factory=collections.Counter)

    def resolve(self, name):
        scope = self
        while scope is not None:
            if name in scope.ns:
                return scope
            scope = scope.parent

    def clone(self):
        return __class__(self.node, self.parent, self.ns.copy())

    def merge_with_branched_scope(self, other):
        '''
        This is used to combine the scopes generated by analyzing each branch
        in an if/else statement. Scopes are combined by taking the union of all
        keys, and returning the maximum number of writes across the branches.
        This ensures that we track all variables that could have been set, and
        are tracking the maximum number of times they were set.
        '''
        assert self.node == other.node and self.parent == other.parent
        return __class__(self.node, self.parent, collections.Counter({
            # Return the maximum number of writes
            key: max(self.ns[key], other.ns[key])
            # Take the union of keys
            for key in self.ns.keys() | other.ns.keys()
        }))

class ClosureScopeAnalysis(ast.NodeVisitor):
    '''
    This analysis enforces some constraints on scopes that are necessary for
    our CPS transform.

    1. Only the local scope can be mutated. Enforced by prohibiting nonlocal and global keywords.
    2. If a nonlocal variable is loaded, we require that it be immutable in its original scope.

    We implement this by doing two passes over the input AST. The first establishes the scope
    associated with each variable reference. The second validates that variables follow rule #2 above.
    Two passes are necessary because names are assigned to scopes statically based on whether it is
    ever possible for them to be defined in that scope.
    '''

    def_pass = 'def_pass'
    use_pass = 'use_pass'

    def __call__(self, node, source):
        self.node = node
        self.source = source
        self.scopes: list[Scope] = []
        self.scope_map = {}
        '''
        We analyze code in two passes. The first pass tracks all stores to
        variables in order to establish scopes. We hold onto those fully
        analyzed scopes in `complete_scope_map`.
        '''
        self.analysis_pass = __class__.def_pass
        with self.in_scope(node):
            self.visit(node)
        self.complete_scope_map = dict(self.scope_map)
        '''
        The main difference in this second pass is that we validate loaded
        variables against statically-determined scopes to ensure closure
        vars are immutable. We still keep the definition machinery to validate
        that variables are defined before closures, in order to avoid issues
        with CPS.
        '''
        self.analysis_pass = __class__.use_pass
        with self.in_scope(node):
            self.visit(node)

    def analyze_scope(self, node):
        '''
        This returns a mapping from scope-defining nodes to a `Scope`.
        '''
        # HACK: Copying above invocation.
        self.scopes: list[Scope] = []
        self.scope_map = {}
        self.analysis_pass = __class__.def_pass
        with self.in_scope(node):
            self.visit(node)
        return self.scope_map

    def generic_visit(self, node):
        # Refactored from ASTVisitor
        for field, value in ast.iter_fields(node):
            self.generic_visit_field(value)

    def generic_visit_field(self, value):
        # Refactored from ASTVisitor
        if isinstance(value, list):
            for item in value:
                if isinstance(item, ast.AST):
                    self.visit(item)
        elif isinstance(value, ast.AST):
            self.visit(value)

    def visit_Lambda(self, node):
        with self.in_scope(node):
            return self.generic_visit(node)

    def visit_FunctionDef(self, node):
        self.add_name_to_scope(node.name)
        self.generic_visit_field(node.decorator_list)
        with self.in_scope(node):
            self.generic_visit_field(node.args)
            self.generic_visit_field(node.body)

    @contextlib.contextmanager
    def in_scope(self, node):
        self.scopes.append(Scope(
            node=node,
            parent=self.current_scope if len(self.scopes) else None,
        ))
        self.scope_map[node] = self.current_scope
        try:
            yield
        finally:
            self.scopes.pop()

    def add_name_to_scope(self, name):
        self.current_scope.ns[name] += 1

    @property
    def current_scope(self):
        return self.scopes[-1]

    def visit_Assign(self, node):
        # We make sure to visit values first.
        self.generic_visit_field(node.value)
        # Then we visit targets.
        self.generic_visit_field(node.targets)

    def visit_AnnAssign(self, node):
        # Value first then target, like above.
        self.generic_visit_field(node.value)
        self.generic_visit_field(node.target)

    def visit_AugAssign(self, node):
        # Value first then target, like above.
        # NOTE: We avoid separately analying a Load() to the target because
        # the case for Load() in visit_Name checks for nonlocals that were mutated
        # but we know this variable is local (because the nonlocal scope can't be changed).
        # So, there's no way a Load() here could detect an issue.
        self.generic_visit_field(node.value)
        self.generic_visit_field(node.target)

    def visit_arg(self, node):
        self.add_name_to_scope(node.arg)

    # Handling comprehensions
    def visit_ListComp(self, node):
        with self.in_scope(node):
            self.generic_visit_field(node.generators)
            self.generic_visit_field(node.elt)
    visit_SetComp = visit_ListComp
    def visit_DictComp(self, node):
        with self.in_scope(node):
            self.generic_visit_field(node.generators)
            self.generic_visit_field(node.key)
            self.generic_visit_field(node.value)
    def visit_comprehension(self, node):
        self.generic_visit_field(node.iter)
        self.generic_visit_field(node.target)
        self.generic_visit_field(node.ifs)

    def visit_If(self, node):
        def _replace_current_scope(scope):
            prev_scope = self.current_scope
            # We want to make sure we're replacing a comparable scope, since we
            # only use this function to change the namespace, but still want
            # to ensure they are associated with the same node and parent scope.
            assert (
                prev_scope.node == scope.node and
                prev_scope.parent == scope.parent and
                prev_scope == self.scope_map[scope.node]
            ), 'Ensures that new scope can replace current scope.'
            self.scopes[-1] = scope
            self.scope_map[scope.node] = scope
            return prev_scope

        # 1) Analyze the test.
        self.generic_visit_field(node.test)

        # 2) Analyze the body.
        # Copy starting scope
        starting = self.current_scope.clone()
        # Using original starting scope, analyze body
        self.generic_visit_field(node.body)

        # 3) Analyze the orelse.
        # Replace scope with starting scope and analyze orelse
        body_scope = _replace_current_scope(starting)
        self.generic_visit_field(node.orelse)

        # 4) Combine the scopes from each branch.
        _replace_current_scope(self.current_scope.merge_with_branched_scope(body_scope))

    # Case for handling names
    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            self.add_name_to_scope(node.id)

        elif self.analysis_pass == __class__.use_pass:
            if isinstance(node.ctx, ast.Load):
                local_scope = self.complete_scope_map[self.current_scope.node]
                scope = local_scope.resolve(node.id)
                is_global = scope is None
                is_local = scope == local_scope
                if not is_global and not is_local:
                    write_count = scope.ns[node.id]
                    if write_count > 1:
                        raise _error_at_node(f'Variable from outer scope must be immutable, but is set in {write_count} locations.', node, self.source)
                    if node.id not in self.scope_map[scope.node].ns:
                        raise _error_at_node('Variable from outer scope must be defined before being referenced.', node, self.source)

    # We block unhandled cases that change scope or bind names.
    # https://realpython.com/python-scope-legb-rule/

    def not_implemented(self, node):
        raise NotImplementedError()

    # Blocked because they make outer scopes mutable.
    visit_Global = not_implemented
    visit_Nonlocal = not_implemented

    # Blocked because they bind names.
    visit_Try = not_implemented
    visit_TryStar = not_implemented
    visit_NamedExpr = not_implemented

    # Blocked because they cause new scopes.
    visit_ClassDef = not_implemented
    visit_GeneratorExp = not_implemented

class NodeTransformer(ast.NodeTransformer):
    def generic_visit(self, node):
        for field, _ in ast.iter_fields(node):
            self.generic_visit_field(node, field)
        return node

    def generic_visit_field(self, node, field):
        # This is the inner loop over fields implemented by `ast.NodeTransformer.generic_visit`.
        # It's useful when a specific order over fields needs to be guaranteed, so we refactor
        # it out here.
        old_value = getattr(node, field)
        if isinstance(old_value, list):
            # This case and next one are both copied from original.
            new_values = []
            for value in old_value:
                if isinstance(value, ast.AST):
                    value = self.visit(value)
                    if value is None:
                        continue
                    elif not isinstance(value, ast.AST):
                        new_values.extend(value)
                        continue
                new_values.append(value)
            old_value[:] = new_values
        elif isinstance(old_value, ast.AST):
            new_node = self.visit(old_value)
            if new_node is None:
                delattr(node, field)
            else:
                setattr(node, field, new_node)

def _error_at_node(msg, node, source, *, filename='tmp.py'):
    # Subtract 1 b/c node.lineno is 1-indexed
    code = source.splitlines()[node.lineno - 1]
    # Add 1 b/c node.col_offset is 0-indexed
    offset = node.col_offset + 1
    return SyntaxError(msg, (filename, node.lineno, offset, code))

class PythonSubsetValidator(ast.NodeVisitor):
    def __call__(self, node, source):
        self.in_loop_flag = False
        self.errors = []
        self.visit(node)
        if self.errors:
            # HACK For now, we just report the first error.
            node = self.errors[0]
            raise _error_at_node('Found unsupported Python feature.', node, source)

    def add_error(self, node):
        self.errors.append(node)

    def visit_error(self, node):
        self.add_error(node)
        super().generic_visit(node)

    visit_Global = visit_error
    visit_Nonlocal = visit_error

    visit_ClassDef = visit_error
    visit_AsyncFunctionDef = visit_error

    visit_AsyncFor = visit_error
    visit_AsyncWith = visit_error
    visit_With = visit_error

    visit_Match = visit_error

    visit_Try = visit_error
    visit_TryStar = visit_error

    visit_Import = visit_error
    visit_ImportFrom = visit_error

    visit_Await = visit_error
    visit_Yield = visit_error
    visit_YieldFrom = visit_error
    visit_NamedExpr = visit_error

    # We prohibit del/set of attributes and subscripts
    def visit_Attribute(self, node):
        super().generic_visit(node)
        if isinstance(node.ctx, (ast.Store, ast.Del)):
            self.add_error(node)
    visit_Subscript = visit_Attribute

    # We prohibit nested functions in loops
    def visit_While(self, node):
        prev_flag = self.in_loop_flag
        self.in_loop_flag = True
        super().generic_visit(node)
        self.in_loop_flag = prev_flag
    visit_For = visit_While

    def visit_Lambda(self, node):
        super().generic_visit(node)
        if self.in_loop_flag:
            self.add_error(node)
    visit_FunctionDef = visit_Lambda

class DesugaringTransform(NodeTransformer):
    """
    This "desugars" the AST by
    - separating out function calls to individual assignments
    - converting <exp> if <cond> else <exp> statements to explicit
    if-else blocks
    - converting lambda expressions to function def
    - converting logical and/or statements to equivalent if/else blocks
    - make None function return explicit
    - converting branching expressions into function calls

    Currently not implemented desugaring:
    - de-decorating function definitions
    """
    temporary_var_prefix = '__v'
    def __call__(self, rootnode):
        self.new_stmt_stack = [[],]
        self.in_expression = False
        self.n_temporary_vars = 0
        self.visit(rootnode)
        rootnode = ast.parse(ast.unparse(rootnode)) #HACK: this might be slow
        return rootnode

    def visit(self, node):
        if isinstance(node, ast.stmt):
            method = "visit_stmt"
        elif (
            isinstance(node, ast.expr) and \
            self.is_branching_expr(node) and \
            not self.is_assigned_to(node)
        ):
            method = "visit_expr"
        else:
            method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_stmt(self, node):
        self.new_stmt_stack.append([])
        node = ast.NodeTransformer.visit(self, node)
        if node is not None:
            self.add_statement(node)
        return self.new_stmt_stack.pop()

    def visit_expr(self, node):
        """
        Since expressions won't change values in the namespace
        (aside from side-effects), we
        wrap *outermost* expression into a thunk to
        avoid needing to branch in the current function.
        """
        if not self.in_expression:
            self.in_expression = True
            thunk_node = ast.parse("(lambda : __expr__)()").body[0].value
            thunk_node.func.body = node
            node = ast.NodeTransformer.visit(self, thunk_node)
            self.in_expression = False
        else:
            node = ast.NodeTransformer.visit(self, node)
        return node

    @classmethod
    def is_assigned_to(cls, node):
        return hasattr(node, "ctx") and isinstance(node.ctx, ast.Store)


    @classmethod
    def is_branching_expr(cls, node):
        if isinstance(node, (ast.IfExp, ast.BoolOp)):
            return True
        for child in ast.iter_child_nodes(node):
            if cls.is_branching_expr(child):
                return True
        return False

    def generate_name(self):
        self.n_temporary_vars += 1
        return f"{DesugaringTransform.temporary_var_prefix}{self.n_temporary_vars - 1}"

    # def visit_Call(self, node):
    #     node = ast.NodeTransformer.generic_visit(self, node)
    #     return_name = self.generate_name()
    #     assn_node = ast.Assign(
    #         targets=[ast.Name(id=return_name, ctx=ast.Store())],
    #         value=node,
    #         lineno=0
    #     )
    #     self.add_statement(assn_node)
    #     return ast.Name(id=return_name, ctx=ast.Load())

    def visit_IfExp(self, node):
        """
        Convert <exp> if <cond> else <exp> to explicit if then else blocks
        """
        return self.desugar_to_IfElse_block(
            test_expr=node.test,
            if_expr=node.body,
            else_expr=node.orelse,
            test_name = self.generate_name(),
            return_name = self.generate_name()
        )

    def visit_Lambda(self, node):
        def_name = self.generate_name()
        def_node = ast.parse(textwrap.dedent(f"""
        def {def_name}():
            return None
        """)).body[0]
        def_node.args = node.args
        def_node.body[0].value = node.body
        def_node = ast.NodeTransformer.generic_visit(self, def_node)
        self.add_statement(def_node)
        return ast.Name(id=def_name, ctx=ast.Load())

    def visit_Return(self, node):
        if node.value is None:
            node.value = ast.Constant(value=None)
        node = self.generic_visit(node)
        return node

    def visit_BoolOp(self, node):
        if len(node.values) > 2:
            leftmost = ast.BoolOp(
                op=type(node.op)(),
                values=node.values[:-1]
            )
            node.values = [leftmost, node.values[-1]]
            return self.visit_BoolOp(node)
        test_name = self.generate_name()
        return_name = self.generate_name()
        if isinstance(node.op, ast.And):
            return self.desugar_to_IfElse_block(
                test_expr=node.values[0],
                if_expr=node.values[1],
                else_expr=ast.Name(id=test_name),
                test_name=test_name,
                return_name=return_name
            )
        elif isinstance(node.op, ast.Or):
            return self.desugar_to_IfElse_block(
                test_expr=node.values[0],
                if_expr=ast.Name(id=test_name),
                else_expr=node.values[1],
                test_name=test_name,
                return_name=return_name
            )
        raise ValueError("BoolOp is neither And nor Or")

    def visit_FunctionDef(self, node):
        node = self.generic_visit(node)
        # make return value of None function explicit
        # Note: this is required for CPSTransform to work
        if not isinstance(node.body[-1], ast.Return):
            return_none = ast.parse("return None").body[0]
            node.body.append(return_none)
        return node

    # def de_decorate_FunctionDef(self, node):
    #     decorator_list, node.decorator_list = node.decorator_list, []
    #     new_block = [self.visit(node)]
    #     for decorator in decorator_list[::-1]:
    #         decorator_stmt = ast.parse(f"{node.name} = __dec__({node.name})").body[0]
    #         decorator_stmt.value.func = decorator
    #         decorator_stmt = self.visit(decorator_stmt)
    #         new_block.append(decorator_stmt)
    #     return new_block

    def visit_comprehension(self, node):
        '''
        Convert comprehensions into recursive_reduce calls. Later generators correspond to
        nested recursive_reduce calls. For the following example input:

        ```
        [
            (x, y)
            for x in range(4)
            if x % 2 == 0
            for y in range(5)
            if x + y < 3
        ]
        ```

        We transform code into the following:

        ```
        def __v0(_acc, x):
            def __v1(_acc, y):
                return __acc + [(x, y)] if all([x + y < 3]) else __acc
            return __acc + recursive_reduce(__v1, range(5), []) if all([x % 2 == 0]) else __acc
        recursive_reduce(__v0, range(4), [])
        ```

        This function can also handle set and dictionary comprehensions.
        '''
        self.add_statement(ast.parse('from flippy import recursive_reduce').body[0])

        nested_cb = None
        if isinstance(node, ast.ListComp):
            op = '+'
            initial = '[]'
            nested_comp = ast.List(elts=[node.elt])
        elif isinstance(node, ast.SetComp):
            op = '|'
            initial = 'set()'
            nested_comp = ast.Set(elts=[node.elt])
        else:
            assert isinstance(node, ast.DictComp)
            op = '|'
            initial = '{}'
            nested_comp = ast.Dict(keys=[node.key], values=[node.value])

        for g in node.generators[::-1]:
            # We make sure to specialize the function body based on whether tests were provided.
            base_expr = f'''__acc {op} {Placeholder.new('nested_comp')}'''
            if len(g.ifs) == 0:
                test = None
                expr = base_expr
            else:
                # We separately handle the common case (one test) from the less common case of multiple tests.
                if len(g.ifs) == 1:
                    test = g.ifs[0]
                else:
                    test = ast.BoolOp(op=ast.And(), values=g.ifs)
                expr = f'''{base_expr} if {Placeholder.new('test')} else __acc'''

            # We add a nested callback so that we can destructure more complex targets. This is particularly useful
            # for dictionary comprehensions. Previously, we only had nested lambdas, which prevented supporting complex targets.
            cb_name = self.generate_name()
            cb = Placeholder.fill(ast.parse(textwrap.dedent(f'''
            def {cb_name}(__acc, __target):
                {Placeholder.new('target')} = __target
                {Placeholder.new('nested_cb')}
                return {expr}
            ''')), dict(
                # In the innermost loop, this CB is None, so the expression is just deleted.
                nested_cb=nested_cb,
                nested_comp=nested_comp,
                test=test,
                target=g.target
            )).body[0]

            new_node = Placeholder.fill(ast.parse(textwrap.dedent(f'''
            recursive_reduce({cb_name}, {Placeholder.new('iter')}, {initial})
            ''')), dict(
                iter=g.iter,
            )).body[0].value

            nested_cb = cb
            nested_comp = new_node

        # We need to add the outermost CB as a statement to the program.
        self.add_statement(self.visit(cb))

        return self.visit(new_node)

    visit_ListComp = visit_comprehension
    visit_SetComp = visit_comprehension
    visit_DictComp = visit_comprehension

    def visit_AugAssign(self, node):
        loadable_target = copy.copy(node.target)
        loadable_target.ctx = ast.Load()

        return self.visit(ast.Assign(
            targets=[node.target],
            value=ast.BinOp(
                left=loadable_target,
                op=node.op,
                right=node.value,
            ),
            lineno=node.lineno,
        ))

    def visit_AnnAssign(self, node):
        if node.value is None:
            # In this case, it's just a type annotation.
            return None
        return self.visit(ast.Assign(
            targets=[node.target],
            value=node.value,
            lineno=node.lineno,
        ))

    def visit_While(self, node):
        '''
        For simplicity, we desugar while loops to make tests be explicit if statements, mirroring how we handle
        other conditional control flow (like if expressions, logical and, and logical or).
        '''
        loop, = Placeholder.fill_from_source(f'''
        while True:
            if not ({Placeholder.new('test')}):
                break
            {Placeholder.new('body')}
        ''', dict(
            test=node.test,
            body=node.body,
        )).body
        # Doing this out of completeness, but it's a constant so nothing will change.
        self.generic_visit_field(loop, 'test')
        self.generic_visit_field(loop, 'body')
        return loop

    def visit_For(self, node):
        '''
        We implement a CPS-compatible `for` by assuming a loop argument that can be indexed and has a length.
        '''

        assert not node.orelse

        iter_sym = f'_for_iter_{node.lineno}'
        idx_sym = f'_for_idx_{node.lineno}'
        code = Placeholder.fill(ast.parse(textwrap.dedent(f'''
        {iter_sym} = {Placeholder.new('iter')}
        {idx_sym} = 0
        while {idx_sym} < len({iter_sym}):
            {Placeholder.new('target')} = {iter_sym}[{idx_sym}]
            {idx_sym} += 1
            {Placeholder.new('body')}
        ''')), dict(
            iter=node.iter,
            target=node.target,
            body=node.body,
        )).body

        return [
            self.visit(s)
            for s in code
        ]

    def desugar_to_IfElse_block(
        self,
        test_expr,
        if_expr,
        else_expr,
        test_name,
        return_name
    ):
        test_node, if_node = ast.parse(textwrap.dedent(f"""
        {test_name} = test
        if {test_name}:
            {return_name} = if_body
        else:
            {return_name} = else_body
        """)).body
        test_node.value = test_expr
        test_node = self.generic_visit(test_node)
        if not isinstance(test_node, list):
            test_node = [test_node]
        if_node.body[0].value = if_expr
        if_node.orelse[0].value = else_expr
        self.generic_visit(if_node)
        for stmt in [*test_node, if_node]:
            self.add_statement(stmt)
        return ast.Name(id=return_name, ctx=ast.Load())

    def add_statement(self, node):
        self.new_stmt_stack[-1].append(node)

class SetLineNumbers(ast.NodeTransformer):
    """
    Reset line numbers by statement.
    """
    def __call__(self, node, starting_line=0):
        self.cur_line = starting_line
        self.visit(node)
        ast.fix_missing_locations(node)
        return node
    def set_line(self, node):
        node.lineno = self.cur_line
        self.cur_line += 1
    def generic_visit(self, node):
        if isinstance(node, (ast.Expr, ast.stmt)):
            self.set_line(node)
        elif hasattr(node, 'lineno'):
            del node.lineno
        ast.NodeTransformer.generic_visit(self, node)
        return node

class GetLineNumber(ast.NodeVisitor):
    """
    Reset line numbers by statement.
    """
    def __call__(self, node: ast.AST, lineno: int):
        self.cur_line = 0
        self.lineno = lineno
        try:
            self.visit(node)
        except StopIteration as err:
            return err.value
        raise ValueError(f"Could not find line number {lineno} in {ast.dump(node)}")
    def set_line(self, node):
        node.lineno = self.cur_line
        self.cur_line += 1
    def generic_visit(self, node):
        if isinstance(node, (ast.Expr, ast.stmt)):
            if self.cur_line == self.lineno:
                raise StopIteration(node)
            self.cur_line += 1
        ast.NodeVisitor.generic_visit(self, node)
        return node

class CPSFunction:
    """
    This class is used to represent a function that has been transformed into
    continuation-passing style (CPS). It wraps a callable and stores the source code.
    It also provides a closure property that contains the variables captured by the function.
    The identity of a CPSFunction is determined by its source code and closure variables
    (see `__hash__` and `__eq__`).

    Note that the behavior of a function could depend on the contents of `global_store`,
    in which case the function is not a pure function. Inference algorithms are
    expected to be aware of this and handle the combination of the function and
    `global_store` appropriately.

    Separately, the behavior of a function could depend on variables defined in
    the global scope of the Python interpreter. In this case, it is assumed
    that these variables are not changed during the execution of the function.
    """
    def __init__(
        self,
        cps_func: Callable,
        func_src: str,
    ):
        self.cps_func = cps_func
        self.func_src = func_src
        setattr(self, CPSTransform.is_transformed_property, True)

    def __call__(self, *args, **kwargs):
        return self.cps_func(*args, **kwargs)

    @property
    def closure(self) -> hashabledict:
        if getattr(self.cps_func, "__closure__", None) is not None:
            closure_keys = self.cps_func.__code__.co_freevars
            closure_values = [cell.cell_contents for cell in self.cps_func.__closure__]
            return hashabledict(zip(closure_keys, closure_values))
        else:
            return hashabledict()

    @cached_property
    def _hash(self):
        # we sort the closure and use repr to avoid circular references
        return hash((self.func_src, repr(sorted(self.closure.items()))))

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, CPSFunction):
            return False
        return \
            self.func_src == other.func_src and \
            self.closure == other.closure

    @property
    def __name__(self):
        return self.cps_func.__name__

class CPSTransform(NodeTransformer):
    """
    Convert python to a form of continuation passing style.
    """
    is_transformed_property = "_cps_transformed"
    func_src_name = "__func_src"
    cps_interpreter_name = "_cps"
    final_continuation_name = "_cont"
    stack_name = "_stack"

    def __init__(self):
        self.current_continuation_stmts = None
        self.parent_function_lineno = 0
        self.continue_break_replacement = None
        self.call_count = 0

    def __call__(self, node):
        self.scope_map = ClosureScopeAnalysis().analyze_scope(node)
        self.scope = None
        self.call_count = 0
        with self.in_scope(node):
            return self.visit(node)

    @classmethod
    def is_transformed(cls, func):
        return getattr(func, CPSTransform.is_transformed_property, False)

    def visit_Lambda(self, node):
        # Most lambdas are desugared, so this only has to handle the ones we generate,
        # like the default value for _cont of lambda val: val.
        with self.in_scope(node):
            return self.generic_visit(node)

    def visit_Module(self, node):
        # If we're in the outermost module scope, we only transform function definitions
        # Everything else is executed as normal (deterministic) python
        # This is primarily to handle cases where non-function definition statements
        # are created in the course of desugaring but still need to be executed.
        for stmt in node.body:
            if isinstance(stmt, ast.FunctionDef):
                self.visit(stmt)
        return node

    def visit_FunctionDef(self, node):
        node = self.add_function_src_to_FunctionDef_body(node)
        with self.in_scope(node):
            self.visit(node.args)
            # we track this so that line numbers passed to interpreter are
            # with respect to the start of the function definition
            old_parent_lineno = self.parent_function_lineno
            self.parent_function_lineno = node.lineno
            node.body, _ = self.transform_block(node.body)
            self.parent_function_lineno = old_parent_lineno

        if hasattr(node, 'body_imports'):
            for i in node.body_imports:
                node.body.insert(1, i) # 1 to make them after function source
            del node.body_imports

        node = self.add_keyword_to_FunctionDef(node, self.stack_name, "()")
        node = self.add_keyword_to_FunctionDef(node, self.cps_interpreter_name, self.cps_interpreter_name)
        node = self.add_keyword_to_FunctionDef(node, self.final_continuation_name, "lambda val: val")

        # decorator = ast.parse(f"lambda fn: (fn, setattr(fn, '{self.is_transformed_property}', True))[0]").body[0].value

        # This decorator position makes it the last decorator called.
        # This ensures that the outermost function has `is_transformed_property` set.
        # node.decorator_list.insert(0, decorator)
        # This decorator position makes it the first decorator called.
        # This is important so that intermediate decorators can change their behavior
        # the second time that this function is executed.
        # Functions are executed a second time when being evaluted after being CPS-transformed
        # because decorators aren't removed by the transform.
        # node.decorator_list.append(decorator)
        return node

    def add_keyword_to_FunctionDef(self, node, key, value):
        assert isinstance(node, ast.FunctionDef)
        cur_keywords = [kw.arg for kw in node.args.kwonlyargs + node.args.args]
        assert key not in cur_keywords
        node.args.kwonlyargs.append(ast.arg(arg=key))
        node.args.kw_defaults.append(ast.parse(value).body[0].value)
        return node

    def add_function_src_to_FunctionDef_body(self, node):
        assert isinstance(node, ast.FunctionDef)
        ctx_id = repr(ast.unparse(node))
        ctx_id_assn = ast.parse(f"{self.func_src_name} = {ctx_id}").body[0]
        node.body.insert(0, ctx_id_assn)
        node.decorator_list.append(
            ast.parse(f"lambda fn: CPSFunction(fn, {ctx_id})").body[0].value
        )
        return node

    @contextlib.contextmanager
    def in_scope(self, node):
        prev_scope = self.scope
        self.scope = self.scope_map[node]
        try:
            yield
        finally:
            self.scope = prev_scope

    def scope_packing_for_current_scope(self, locals_name, scope_name, *, hashabledict=False):
        '''
        Our general scheme for scope packing is to pack the possibly-present names
        into a dictionary, if they are defined. Variables are unpacked in a way that
        handles the issue of definition, by only conditionally defining variables.

        Note that this formerly tracked names as they were defined in the local scope,
        but this was removed in favor of simply adding all names that could possibly be in
        the present scope. We did this because loops make it more difficult to reason about
        variables, requiring some care to ensure that nested continuation-creating nodes
        would propagate variables that would be defined later in the loop body.
        '''
        names = sorted(self.scope.ns.keys())
        import_ = 'from flippy.hashable import hashabledict' if hashabledict else ''
        scope_type = 'hashabledict' if hashabledict else ''
        pack = f'''
        {import_}
        # We capture locals before we enter any other scope (like a lambda or list comprehension)
        {locals_name} = locals()
        {scope_name} = {scope_type}({{name: {locals_name}[name] for name in {names} if name in {locals_name}}})
        '''
        unpack = '\n'.join(f'''
        if "{name}" in {scope_name}:
            {name} = {scope_name}["{name}"]
        ''' for name in names) or 'pass'
        return (
            ast.parse(textwrap.dedent(pack)).body,
            ast.parse(textwrap.dedent(unpack)).body,
        )

    def visit_Assign(self, node):
        # We make sure to visit values first.
        self.generic_visit_field(node, 'value')
        # Then we visit targets.
        self.generic_visit_field(node, 'targets')
        return node

    def visit_Call(self, node):
        '''
        Replace the Call node with a continuation and call to the interpreter
        that picks up at the continuation.
        '''
        call_id = self.call_count
        self.call_count += 1
        continuation_name = f"_cont_{call_id}"
        result_name = f"_res_{call_id}"
        locals_name = f"_locals_{call_id}" #NOTE: could optimize with _locals_{node.lineno}
        pack, unpack = self.scope_packing_for_current_scope(locals_name, f"_scope_{call_id}")
        code = Placeholder.fill(ast.parse(textwrap.dedent(f'''
            {Placeholder.new('pack')}
            def {continuation_name}({result_name}):
                {Placeholder.new('unpack')}
            return lambda : {self.cps_interpreter_name}.interpret(
                {Placeholder.new('func')},
                cont={continuation_name},
                stack={self.stack_name},
                func_src={self.func_src_name},
                locals_={locals_name},
                call_id={call_id},
                lineno={node.lineno - self.parent_function_lineno}
            )()
        ''')), dict(
            func=self.visit(node.func),
            pack=pack,
            unpack=unpack,
        )).body
        continuation_node, thunk_node = code[-2:]
        thunk_node.value.body.args = [self.visit(arg) for arg in node.args]
        thunk_node.value.body.keywords = [self.visit(kw) for kw in node.keywords]
        self.current_continuation_stmts.extend(code)
        # Update CCS to the continuation of our function.
        self.current_continuation_stmts = continuation_node.body
        return ast.Name(id=result_name, ctx=ast.Load())

    def visit_While(self, node):
        '''
        We transform While nodes into a CPS style, by having
        1) a conditionally-recursive (dependent on loop test) continuation that executes the loop body and
        2) a final continuation, visited on loop break or False value from loop test.
        We have to carefully pack and unpack scope through all these functions. The current implementation
        handles temporary variables and changes to scope before a continue/break.
        '''
        # Store previous continue/break replacements, for recursive processing
        previous = self.continue_break_replacement

        call_id = self.call_count
        self.call_count += 1

        loop_name = f"_loop_fn_{call_id}"
        loop_exit_name = f"_loop_exit_fn_{call_id}"
        scope_name = f"_scope_{call_id}"
        locals_name = f"_locals_{call_id}"

        # We process body, replacing continue/break with placeholders.
        # We initially replace with placeholders because continue/break requires packing any
        # variables that could be defined in loop body. Once we know the variables that
        # need to be packed, we replace with the appropriate calls (below).
        self.continue_break_replacement = (
            ast.Expr(ast.Name(Placeholder.new('continue_'))),
            ast.Expr(ast.Name(Placeholder.new('break_'))))
        node_body, continuation_block = self.transform_block(node.body)

        # Now that we've transformed the loop body, we know all variables that can be defined.
        pack, unpack = self.scope_packing_for_current_scope(locals_name, scope_name, hashabledict=True)
        def loop_call_template(cont):
            return textwrap.dedent(f'''
                return lambda: {self.cps_interpreter_name}.interpret(
                    {loop_name},
                    cont={cont},
                    stack={self.stack_name},
                    func_src={self.func_src_name},
                    locals_={locals_name},
                    lineno={node.lineno - self.parent_function_lineno}
                )({scope_name})''')
        # For the loop's initial call, we supply the post-loop continuation. However, for the
        # recursive calls (at end of loop body or for continue statements), we use final_continuation_name
        loop_init_call = pack + ast.parse(loop_call_template(loop_exit_name)).body
        loop_recur_call = pack + ast.parse(loop_call_template(self.final_continuation_name)).body
        loop_exit_call = pack + ast.parse(f'return lambda: {self.final_continuation_name}(_ValueWrapper({scope_name}))').body
        # At the end of body execution, we recurse. Must happen before filling placeholders in node_body.
        continuation_block.extend(loop_recur_call)
        # Replace continue/break with call that has appropriate variable packing.
        node_body = [Placeholder.fill(s, dict(continue_=loop_recur_call, break_=loop_exit_call)) for s in node_body]

        # We import _ValueWrapper at the beginning of our enclosing scope.
        assert isinstance(self.scope.node, ast.FunctionDef)
        self.scope.node.body_imports = ast.parse('from flippy.transforms import _ValueWrapper').body

        code = Placeholder.fill(ast.parse(textwrap.dedent(f'''
            def {loop_name}({scope_name}):
                {Placeholder.new('unpack')}
                if {Placeholder.new('test')}:
                    {Placeholder.new('body')}
                else:
                    # When exiting, we have no further assignments, so we could reuse initial scope, but don't.
                    {Placeholder.new('loop_exit_call')}
            def {loop_exit_name}({scope_name}):
                if not isinstance({scope_name}, _ValueWrapper):
                    return lambda: {self.final_continuation_name}({scope_name})
                {scope_name} = {scope_name}.value
                {Placeholder.new('unpack')}
            {Placeholder.new('loop_init_call')}
        ''')), dict(
            # NOTE: We assume the above desugaring, so that the loop test is trivial (always True), and our loop predicate
            # is instead being tested with an if statement in the loop's body.
            test=node.test,
            body=node_body,
            unpack=unpack,
            loop_init_call=loop_init_call,
            loop_recur_call=loop_recur_call,
            loop_exit_call=loop_exit_call,
        )).body
        loop_fn, loop_exit_fn = code[:2]

        # Next few lines are copied from visit_FunctionDef
        loop_fn = self.add_function_src_to_FunctionDef_body(loop_fn)
        loop_fn = self.add_keyword_to_FunctionDef(loop_fn, self.stack_name, "()")
        loop_fn = self.add_keyword_to_FunctionDef(loop_fn, self.cps_interpreter_name, self.cps_interpreter_name)
        loop_fn = self.add_keyword_to_FunctionDef(loop_fn, self.final_continuation_name, "lambda val: val")
        assert loop_fn is code[0], 'Making sure the object was mutated'

        self.current_continuation_stmts.extend(code)
        # Execution continues in the loop exit continuation.
        self.current_continuation_stmts = loop_exit_fn.body
        # Restore, for recursive processing
        self.continue_break_replacement = previous
        return None

    def visit_Continue(self, node):
        c, b = self.continue_break_replacement
        return c
    def visit_Break(self, node):
        c, b = self.continue_break_replacement
        return b

    def transform_block(self, untransformed_stmts):
        '''
        This function transforms each statement in the argument `untransformed_stmts`, in order,
        into CPS. It maintains a list to hold the transformed statements. However,
        in the course of the CPS transform, Call/If/While will produce new continuations.
        So, this algorithm maintains a pointer to the current continuation, updated by
        continuation-generating visitor methods.
        '''

        # Store current state, so we can recursively transform.
        previous = self.current_continuation_stmts

        # We create a block to hold CPS-transformed statements.
        # Initially, these refer to the same list, but self.current_continuation_stmts
        # will change after we process a statement that needs a CPS-transform
        # (i.e. after a Call, While, or If).
        transformed_stmts = self.current_continuation_stmts = []

        for us in untransformed_stmts:
            rv = self.visit(us)
            # We drop return values that are None. This happens in visit_If and visit_While.
            if rv is not None:
                # Note: self.current_continuation_stmts can be modified in other functions,
                # which means this doesn't always append into transformed_stmts.
                if isinstance(rv, list):
                    self.current_continuation_stmts.extend(rv)
                else:
                    self.current_continuation_stmts.append(rv)

        # Hold onto a reference before we restore state.
        ccs = self.current_continuation_stmts
        # Restore previous state.
        self.current_continuation_stmts = previous

        return transformed_stmts, ccs

    def visit_If(self, node):
        """
        We transform each branch of the If node individually, making sure they each wind up
        executing the same continuation, which contains future statements.
        """
        node.test = self.visit(node.test)
        node.body, body_continuation = self.transform_block(node.body)
        node.orelse, orelse_continuation = self.transform_block(node.orelse)
        call_id = self.call_count
        self.call_count += 1

        continuation_name = f"_cont_{call_id}"
        scope_name = f"_scope_{call_id}"
        # Should only do scope packing after variables from both branches have been analyzed.
        pack, unpack = self.scope_packing_for_current_scope(f"_locals_{call_id}", scope_name)
        continuation_node, = Placeholder.fill(ast.parse(textwrap.dedent(f'''
            def {continuation_name}({scope_name}):
                {Placeholder.new('unpack')}
        ''')), dict(unpack=unpack)).body
        conditional_done_call = pack + ast.parse(f'return lambda: {continuation_name}({scope_name})').body
        # Add the continuation
        self.current_continuation_stmts.extend([continuation_node, node])
        # Make branches go to continuation
        body_continuation += conditional_done_call
        orelse_continuation += conditional_done_call
        # Future statements go in continuation body
        self.current_continuation_stmts = continuation_node.body
        return None

    def visit_Return(self, node):
        new_node = ast.parse(f"return lambda : {self.final_continuation_name}(_res)").body[0]
        new_node.value.body.args[0] = self.visit(node.value)
        return new_node

class HashableCollectionTransform(ast.NodeTransformer):
    """
    Wrap all lists and dicts in hashable versions.
    """
    def __call__(self, rootnode):
        self.visit(rootnode)
        return rootnode

    def visit_List(self, node):
        return ast.Call(
            func=ast.Name(id='hashablelist', ctx=ast.Load()),
            args=[node],
            keywords=[]
        )

    def visit_Dict(self, node):
        return ast.Call(
            func=ast.Name(id='hashabledict', ctx=ast.Load()),
            args=[node],
            keywords=[]
        )



"""
FlipPy provides a core API for managing the non-deterministic execution of
probabilistic Python programs in terms of `ProgramState` objects.
The basic model involves intercepting `Distribution.sample` and `Distribution.observe`
calls, represented by `SampleState` and `ObserveState`, respectively. By controlling
execution at these points, we can implement various inference algorithms.

For more background on this approach, see
[van de Meent et al. (2018)](https://arxiv.org/abs/1809.10756).
"""

from typing import Any, Callable, Hashable, Tuple, TYPE_CHECKING, TypeVar, Sequence, Union

from flippy.distributions import Distribution
from flippy.funcutils import cached_property
from flippy.hashable import hashabledict, hashablelist, hashableset
from flippy.types import Continuation, VariableName, ReturnValue

if TYPE_CHECKING:
    from flippy.interpreter import CPSInterpreter, Stack

class ProgramState:
    """
    A program state represents a point in the execution of a program
    (e.g., the current call stack, memory, and line number) that is
    exposed to an external interpreter (e.g., an inference algorithm).
    An external interpreter can control the behavior of a program using the
    `step` method. An interpreter can also access readable attributes and
    meta-data associated with program state.
    Internally, a program state stores a continuation that is used to resume
    execution when the interpreter calls `step`.
    """
    value = None
    def __init__(
        self,
        continuation : 'Continuation' = None,
        name: 'VariableName' = None,
        stack: 'Stack' = None,
        cps : 'CPSInterpreter' = None,
        init_global_store : 'GlobalStore' = None,
    ):
        self.continuation = continuation
        self._name = name
        self.stack = stack
        self.init_global_store = init_global_store
        self.cps = cps

    def set_init_global_store(self, global_store : 'GlobalStore', force=False):
        assert self.init_global_store is None or force, "Cannot set global store twice"
        self.init_global_store = global_store

    def step(self, *args, **kws) -> 'ProgramState':
        """
        Uses a trampoline to execute a sequence of thunks until
        a ProgramState is encountered.
        """
        next_ = self.continuation(*args, **kws)
        global_store = self.init_global_store.copy()
        with self.cps.set_global_store(global_store):
            while True:
                if callable(next_):
                    next_ = next_()
                elif isinstance(next_, ProgramState):
                    next_.set_init_global_store(global_store)
                    return next_
                else:
                    raise TypeError(f"Unknown type {type(next_)}")

    @cached_property
    def name(self) -> 'VariableName':
        if self._name is not None:
            return self._name
        if self.stack is None:
            return None
        return self.stack.without_locals()

    def __eq__(self, other: 'ProgramState'):
        if not isinstance(other, ProgramState):
            return False
        if self._hash != other._hash:
            return False
        return (
            self.__class__ == other.__class__ and
            self.stack == other.stack and
            self.init_global_store.store == other.init_global_store.store
        )

    @cached_property
    def _hash(self):
        return hash((self.__class__, self.stack, hashabledict(self.init_global_store.store), self.value))

    def __hash__(self):
        return self._hash

class ReadOnlyProxy(object):
    def __init__(self):
        self.proxied = None
    def __getattr__(self, name):
        if self.proxied is None:
            raise NotImplementedError("Proxying to None")
        return getattr(self.proxied, name)
    def __contains__(self, key):
        if self.proxied is None:
            raise NotImplementedError("Proxying to None")
        return key in self.proxied
    def __getstate__(self):
        return self.__dict__
    def __setstate__(self, state):
        self.__dict__ = state

class GlobalStore:
    def __init__(self, initial : dict = None):
        self.store = initial if initial is not None else {}

    def copy(self):
        return GlobalStore({**self.store})

    def get(self, key : Hashable, default : Any = None):
        return self.store.get(key, default)

    def __getitem__(self, key : Hashable):
        return self.store[key]

    def __setitem__(self, key : Hashable, value : Any):
        self.store[key] = value

    def __contains__(self, key : Hashable):
        return key in self.store

    def set(self, key : Hashable, value : Any):
        self.store.__setitem__(key, value)

global_store = ReadOnlyProxy()

class InitialState(ProgramState):
    """Represents the initial state of a program."""
    def __init__(
        self,
        continuation : 'Continuation' = None,
        cps : 'CPSInterpreter' = None,
    ):
        super().__init__(
            continuation=continuation,
            cps=cps,
            init_global_store=GlobalStore()
        )

class ObserveState(ProgramState):
    """
    Represents the state of a program when observing a value from a distribution.
    This is the main interface for conditioning.
    """
    def __init__(
        self,
        continuation: 'Continuation',
        distribution: Distribution,
        value: Any,
        name: 'VariableName',
        stack: 'Stack',
        cps : 'CPSInterpreter'
    ):
        super().__init__(
            continuation=continuation,
            name=name,
            stack=stack,
            cps=cps
        )
        self.distribution = distribution
        self.value = value

class SampleState(ProgramState):
    """
    Represents the state of a program when sampling from a `Distribution`.
    """
    def __init__(
        self,
        continuation: 'Continuation',
        distribution: Distribution,
        name: 'VariableName',
        stack: 'Stack',
        cps : 'CPSInterpreter',
        initial_value: Any,
        fit: bool
    ):
        super().__init__(
            continuation=continuation,
            name=name,
            stack=stack,
            cps=cps
        )
        self.distribution = distribution
        self.initial_value = initial_value
        self.fit = fit

class ReturnState(ProgramState):
    """
    Represents the state of a program when a function
    returns a value.
    """
    def __init__(self, value: 'ReturnValue', stack: 'Stack'):
        super().__init__(stack=stack, name="RETURN_STATE")
        if isinstance(value, dict):
            value = hashabledict(value)
        elif isinstance(value, list):
            value = hashablelist(value)
        elif isinstance(value, set):
            value = hashableset(value)
        self.value = value

    def step(self, *args, **kws):
        raise ValueError



import dataclasses
from functools import cached_property
from typing import Mapping, Hashable, Callable, List, Union, TYPE_CHECKING

from flippy.core import ReturnState, SampleState, ObserveState, ProgramState
from flippy.distributions import Categorical, RandomNumberGenerator, Distribution, \
    Dirichlet
from flippy.distributions.random import default_rng

from flippy.types import SampleValue, VariableName

@dataclasses.dataclass
class Entry:
    value : 'SampleValue'
    program_state : Union[SampleState, ObserveState] = None

    @property
    def is_sample(self) -> bool:
        return isinstance(self.program_state, SampleState)

    @cached_property
    def score(self) -> float:
        return self.program_state.distribution.log_probability(self.value)

    @property
    def name(self) -> 'VariableName':
        return self.program_state.name

    @property
    def distribution(self) -> Distribution:
        return self.program_state.distribution

class Trace:
    def __init__(self):
        self._entries : List[Entry] = []
        self._entry_name_order : Mapping['VariableName', int] = {}
        self._return_state : ReturnState = None

    @staticmethod
    def run_from(
        ps : ProgramState,
        old_trace : Union['Trace', None],
        sample_site_callback : Callable[[SampleState], 'SampleValue'],
        observe_site_callback : Callable[[ObserveState], 'SampleValue'],
        break_early : bool = True
    ) -> 'Trace':
        new_trace = Trace()
        if old_trace is not None and len(old_trace) > 0:
            if ps.name not in old_trace:
                raise ValueError(f"Name {ps.name} not already in trace")
            new_trace._entries = old_trace._entries[:old_trace._entry_name_order[ps.name]]
            new_trace._entry_name_order = {
                e.name : i for i, e in enumerate(new_trace._entries)
            }

        while True:
            assert ps.name not in new_trace, f"Name {ps.name} already in trace"
            if isinstance(ps, SampleState):
                value = sample_site_callback(ps)
                new_trace.add_site(ps, value)
                step_args = (value,)
            elif isinstance(ps, ObserveState):
                new_trace.add_site(ps, observe_site_callback(ps))
                step_args = ()
            elif isinstance(ps, ReturnState):
                new_trace.add_return_state(ps)
                break
            if break_early and new_trace._entries[-1].score == float('-inf'):
                break
            ps = ps.step(*step_args)
        return new_trace

    def add_site(
        self,
        program_state : Union[SampleState, ObserveState],
        value : 'SampleValue',
    ):
        self._entries.append(Entry(
            value=value,
            program_state=program_state
        ))
        self._entry_name_order[program_state.name] = len(self._entries) - 1

    def add_return_state(self, program_state : ReturnState):
        self._return_state = program_state

    @property
    def return_value(self):
        assert self._return_state is not None, \
            "Trace does not have a return value"
        return self._return_state.value

    def __len__(self):
        return len(self._entries)

    def __contains__(self, key):
        return key in self._entry_name_order

    @property
    def total_score(self) -> float:
        if self._entries[-1].score == float('-inf'):
            return float('-inf')
        return sum(e.score for e in self._entries)

    def entries(self, start_name : Hashable = None):
        if start_name is None:
            start_idx = 0
        else:
            start_idx = self._entry_name_order[start_name]
        for e in self._entries[start_idx:]:
            yield e

    @property
    def sample_site_names(self) -> List['VariableName']:
        return [e.name for e in self._entries if e.is_sample]

    def __getitem__(self, key):
        return self._entries[self._entry_name_order[key]]



import ast
import textwrap
from flippy.transforms import *
import pytest

def _analyze(src):
    src = textwrap.dedent(src)
    node = ast.parse(src)
    a = ClosureScopeAnalysis()
    a(node, src)
    return a, node

def _analyze_with_expected_error(msg, src):
    with pytest.raises(SyntaxError) as err:
        _analyze(src)
    assert msg in str(err)

def test_scope_analysis():
    # Defined before
    _analyze('''
    def fn():
        x = 3
        def inner():
            return x
        return inner
    ''')

    # Defined after
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            def inner():
                return x
            x = 3
            return inner
        ''')
    assert 'defined before' in str(err)

    # Defined before, nested
    _analyze('''
    def fn():
        x = 3
        def outer():
            def inner():
                return x
            return inner
        outer()
    ''')

    # Defined after, nested
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            def outer():
                def inner():
                    return x
                return inner
            x = 3
            outer()
        ''')
    assert 'defined before' in str(err)

    # Mutating local
    _analyze('''
    def fn():
        def inner():
            x = 3
            x = 4
            return x
        return inner
    ''')

    # Defined before, mutated before
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            x = 3
            x = x + 4
            def inner():
                return x
            return inner
        ''')
    assert 'immutable' in str(err)

    # Defined before, mutated after
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            x = 3
            def inner():
                return x
            x = 4
            return inner
        ''')
    assert 'immutable' in str(err)

    # Defined after, mutated after
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            def inner():
                return x
            x = 3
            x = 4
            return inner
        ''')
    assert 'immutable' in str(err)

def test_scope_analysis_AnnAssign():
    # Defined before
    _analyze('''
    def fn():
        x: int = 3
        def inner():
            return x
        return inner
    ''')

    # Defined after
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            def inner():
                return x
            x: int = 3
            return inner
        ''')
    assert 'defined before' in str(err)

def test_scope_analysis_FunctionDef():
    # Mutating non-local
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            def abc(): pass
            def abc(): pass
            def inner():
                return abc()
            return inner
        ''')
    assert 'immutable' in str(err)

def test_scope_analysis_AugAssign():
    # Mutating local
    _analyze('''
    def fn():
        def inner():
            x = 3
            x += 4
            return x
        return inner
    ''')

    # Mutating non-local
    with pytest.raises(SyntaxError) as err:
        _analyze('''
        def fn():
            x = 3
            x += 4
            def inner():
                return x
            return inner
        ''')
    assert 'immutable' in str(err)

def test_scope_analysis_shadowing():
    _analyze('''
    def fn():
        x = 3
        x = 3
        def inner():
            x = 4
            x = 4
            return x
        return inner() + x
    ''')

def test_scope_analysis_Comp():
    for begin, end, elt in [
        # List comprehension
        ('[', ']', lambda x: x),
        # Set comprehension
        ('{', '}', lambda x: x),
        # Dict comprehension
        ('{', '}', lambda x: f'{x}:{x}'),
    ]:
        # Test that comprehension has a separate scope.
        _analyze(f'''
        def fn():
            x = 3
            y = {begin}{elt('x')} for x in range(3){end}
            def inner():
                return x
            return inner
        ''')

        # Tests when overwriting variables in comprehension
        # NOTE: These are ok because they never have a nested scope.
        _analyze(f'''
        def fn():
            x = [3, 4]
            return {begin}{elt('x')} for x in x{end}
        ''')
        _analyze(f'''
        def fn():
            x = [3, 4]
            return {begin}{elt('x')} for x in x for x in x{end}
        ''')

        # Ensure reference to nonlocal mutated elt raises
        with pytest.raises(SyntaxError) as err:
            _analyze(f'''
            def fn():
                c = [3, 4]
                c = [5]
                def inner():
                    return {begin}{elt('c')} for x in [3, 4]{end}
                return inner
            ''')
        assert 'immutable' in str(err)

        # Ensure reference to nonlocal mutated if raises
        with pytest.raises(SyntaxError) as err:
            _analyze(f'''
            def fn():
                c = [3, 4]
                c = [5]
                def inner():
                    return {begin}{elt('x')} for x in [3, 4] if c{end}
                return inner
            ''')
        assert 'immutable' in str(err)

        # Reference to mutated iter in local scope raises
        # NOTE: This seems conservative, but seems tricky to handle elegantly with
        # multiple comprehensions.
        with pytest.raises(SyntaxError) as err:
            _analyze(f'''
            def fn():
                c = [3, 4]
                c = [5]
                x = {begin}{elt('x')} for x in c{end}
            ''')
        assert 'immutable' in str(err)

        # Ensure reference to nonlocal mutated iter raises
        with pytest.raises(SyntaxError) as err:
            _analyze(f'''
            def fn():
                c = [3, 4]
                c = [5]
                def inner():
                    return {begin}{elt('x')} for x in c{end}
                return inner
            ''')
        assert 'immutable' in str(err)

        # Ensure reference to nonlocal mutated iter raises, for second comprehension
        with pytest.raises(SyntaxError) as err:
            _analyze(f'''
            def fn():
                c = [3, 4]
                c = [5]
                def inner():
                    return {begin}{elt('x')} for a in [1] for x in c{end}
                return inner
            ''')
        assert 'immutable' in str(err)

    # Ensure reference to nonlocal mutated elt raises
    # Special case for dict key VS value
    for elt in ['x: c', 'c: x']:
        with pytest.raises(SyntaxError) as err:
            _analyze(f'''
            def fn():
                c = 0
                c = 1
                def inner(): return {{{elt} for x in []}}
            ''')
        assert 'immutable' in str(err)

def test_scope_analysis_If():
    template = lambda s: f'''
    def fn():
        before = 3
        if True:
            both1 = 1
            body1 = 1
            body2 = 2
            body2 = 3
            body2orelse1 = 4
            body2orelse1 = 5
            body1orelse2 = 7
        else:
            both1 = 2
            orelse1 = 2
            orelse2 = 3
            orelse2 = 4
            body2orelse1 = 7
            body1orelse2 = 8
            body1orelse2 = 9
        after = 8
        def inner():
            {s}
    '''
    a, node = _analyze(template('pass'))
    assert a.complete_scope_map[node.body[0]].ns == collections.Counter(dict(
        both1=1,
        body1=1,
        body2=2,
        orelse1=1,
        orelse2=2,
        body2orelse1=2,
        body1orelse2=2,
        inner=1,
        before=1,
        after=1,
    ))
    # Variables used once, in either or both branches, are ok.
    _analyze(template('return both1'))
    _analyze(template('return body1'))
    _analyze(template('return orelse1'))
    _analyze_with_expected_error('immutable', template('return body2'))
    _analyze_with_expected_error('immutable', template('return orelse2'))
    _analyze_with_expected_error('immutable', template('return body1orelse2'))
    _analyze_with_expected_error('immutable', template('return body2orelse1'))

def test_scope_analysis_For():
    with pytest.raises(SyntaxError) as err:
        _analyze(f'''
        def fn():
            def c(): return idx
            for idx in range(3):
                pass
            return c()
        ''')
    assert 'before being referenced' in str(err)

    # We don't rule this out here -- but we do in the subset analysis
    _analyze(f'''
    def fn():
        for idx in range(3):
            if idx == 0:
                def c(): return idx
        return c()
    ''')



import math
from dataclasses import dataclass
from collections.abc import Iterable
from typing import Sequence, Set, Union, TYPE_CHECKING, Generic, Callable, TypeVar
from itertools import combinations_with_replacement, product
from functools import cached_property

from flippy.tools import isclose, ISCLOSE_RTOL, ISCLOSE_ATOL

if TYPE_CHECKING:
    from flippy.distributions.base import Distribution

Support = Union[
    Sequence,
    Set,
    "Set_"
]

Element = TypeVar('Element')

class Set_(Generic[Element]):
    def __contains__(self, ele: Element) -> bool:
        raise NotImplementedError

    def __pow__(self, exp: int):
        if not isinstance(exp, int) or exp < 0:
            raise ValueError("Exponent must be a non-negative integer")
        return ProductSet(*(self for _ in range(exp)))

@dataclass(frozen=True)
class Interval(Set_):
    start: float
    end: float
    left_open: bool = False
    right_open: bool = False

    def __contains__(self, ele: Element) -> bool:
        if isinstance(ele, (float, int)):
            if self.left_open:
                after_start = ele > self.start
            else:
                after_start = ele >= self.start
            if self.right_open:
                before_end = ele < self.end
            else:
                before_end = ele <= self.end
            return after_start and before_end
        return False

class Range(Set_):
    def __init__(self, *args):
        if len(args) == 1:
            self.start, self.end, self.step = 0, args[0], 1
        elif len(args) == 2:
            self.start, self.end, self.step = args[0], args[1], 1
        elif len(args) == 3:
            self.start, self.end, self.step = args
        else:
            raise ValueError("Range requires 1, 2, or 3 arguments: Range(stop), Range(start, stop), or Range(start, stop, step)")
        assert isinstance(self.start, int), "Start must be an integer"
        assert isinstance(self.end, int) or self.end == float('inf'), "End must be an integer"
        assert isinstance(self.step, int), "Step must be an integer"
        assert self.step > 0, "Step must be positive"

    def __contains__(self, ele: Element) -> bool:
        if not isinstance(ele, int):
            return False
        return (
            self.start <= ele < self.end and
            (ele - self.start) % self.step == 0
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.start}, {self.end}, {self.step})"

class ProductSet(Set_):
    def __init__(self, *seqs):
        self.seqs = seqs
    def __contains__(self, ele: Sequence[Element]) -> bool:
        return (
            isinstance(ele, Iterable) and
            len(ele) == len(self.seqs) and
            all(e in s for e, s in zip(ele, self.seqs))
        )
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(map(repr, self.seqs))})"
    def __iter__(self):
        yield from product(*self.seqs)
    def __len__(self):
        return math.prod([len(s) for s in self.seqs])

class UnionSet(Set_):
    def __init__(self, *sets: Set_):
        self.sets = sets
    def __contains__(self, ele: Element) -> bool:
        return any(ele in s for s in self.sets)
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(map(repr, self.sets))})"

class Simplex(Set_):
    def __init__(self, dimensions):
        self.dimensions = dimensions
    def __contains__(self, ele: Sequence[float]) -> bool:
        return (
            isinstance(ele, (tuple, list)) and \
            len(ele) == self.dimensions and \
            isclose(1.0, sum(ele))
        )
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.dimensions})"


class OrderedIntegerPartitions(Set_):
    # https://en.wikipedia.org/wiki/Composition_(combinatorics)
    def __init__(self, total, partitions):
        self.total = total
        self.partitions = partitions

    def __contains__(self, ele: Sequence[int]) -> bool:
        return (
            isinstance(ele, (tuple, list)) and \
            all(isinstance(x, int) and x >= 0 for x in ele) and \
            len(ele) == self.partitions and \
            sum(ele) == self.total
        )

    @cached_property
    def _enumerated_partitions(self):
        all_partitions = []
        for bins in combinations_with_replacement(range(self.total + 1), self.partitions - 1):
            partition = []
            for left, right in zip((0, ) + bins, bins + (self.total,)):
                partition.append(right - left)
            all_partitions.append(tuple(partition))
        return tuple(all_partitions)

    def __iter__(self):
        yield from self._enumerated_partitions

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(total={self.total}, partitions={self.partitions})"



import collections.abc
from functools import cached_property

"""
Primitive container types that can be hashed.
They are immutable but otherwise behave like normal containers.
The main use case for these is to be used as keys in return distributions
"""

def make_hashable(obj):
    if isinstance(obj, dict):
        return hashabledict(obj)
    elif isinstance(obj, list):
        return hashablelist(obj)
    elif isinstance(obj, set):
        return hashableset(obj)
    elif isinstance(obj, (collections.abc.ItemsView, collections.abc.KeysView)):
        return hashableset(obj)
    elif isinstance(obj, collections.abc.ValuesView):
        return hashablelist(obj)
    return obj

class hashabledict(dict):
    @cached_property
    def _hash(self):
        try:
            return hash(frozenset(self.items()))
        except TypeError:
            return self._recursive_hash()
    def _recursive_hash(self):
        # This only handles cases where the immediate values that are containers
        # can be coerced into hashable versions
        items = ((k, make_hashable(v)) for k, v in self.items())
        return hash(frozenset(items))
    def __hash__(self):
        return self._hash
    def __repr__(self) -> str:
        return super().__repr__()
    def _immutable_error(self, *args, **kwargs):
        raise TypeError(f"{self.__class__.__name__} is immutable; use {self.__class__.__name__}.copy() first")
    def __or__(self, *args, **kwargs):
        return hashabledict(super().__or__(*args, **kwargs))
    def __ror__(self, *args, **kwargs):
        return hashabledict(super().__ror__(*args, **kwargs))
    __setitem__ = _immutable_error
    __delitem__ = _immutable_error
    update = _immutable_error
    clear = _immutable_error
    pop = _immutable_error
    popitem = _immutable_error
    setdefault = _immutable_error
    __ior__ = _immutable_error
    def __reduce__(self):
        return (
            hashabledict,
            (dict(self),)
        )

class hashablelist(list):
    @cached_property
    def _hash(self):
        try:
            return hash(tuple(self))
        except TypeError:
            return self._recursive_hash()
    def _recursive_hash(self):
        # This only handles cases where the immediate values that are containers
        # can be coerced into hashable versions
        items = (make_hashable(v) for v in self)
        return hash(tuple(items))
    def __hash__(self):
        return self._hash
    def __repr__(self) -> str:
        return super().__repr__()
    def _immutable_error(self, *args, **kwargs):
        raise TypeError(f"{self.__class__.__name__} is immutable; use {self.__class__.__name__}.copy() first")
    def __add__(self, *args, **kwargs):
        return hashablelist(super().__add__(*args, **kwargs))
    def __radd__(self, other):
        return self.__add__(other)
    def __mul__(self, *args, **kwargs):
        return hashablelist(super().__mul__(*args, **kwargs))
    def __rmul__(self, other):
        return self.__mul__(other)
    def __getitem__(self, *args, **kwargs):
        if isinstance(args[0], slice):
            return hashablelist(super().__getitem__(*args, **kwargs))
        return super().__getitem__(*args, **kwargs)
    __setitem__ = _immutable_error
    __delitem__ = _immutable_error
    append = _immutable_error
    extend = _immutable_error
    insert = _immutable_error
    pop = _immutable_error
    remove = _immutable_error
    clear = _immutable_error
    __iadd__ = _immutable_error
    sort = _immutable_error
    __imul__ = _immutable_error
    reverse = _immutable_error
    def __reduce__(self):
        return (
            hashablelist,
            (list(self),)
        )

class hashableset(set):
    @cached_property
    def _hash(self):
        return hash(frozenset(self))
    def __hash__(self):
        return self._hash
    def __repr__(self) -> str:
        return repr(set(self))
    def _immutable_error(self, *args, **kwargs):
        raise TypeError(f"{self.__class__.__name__} is immutable; use {self.__class__.__name__}.copy() first")
    def __or__(self, *args, **kwargs):
        return hashableset(super().__or__(*args, **kwargs))
    def __ror__(self, *args, **kwargs):
        return hashableset(super().__ror__(*args, **kwargs))
    def __and__(self, *args, **kwargs):
        return hashableset(super().__and__(*args, **kwargs))
    def __rand__(self, *args, **kwargs):
        return hashableset(super().__rand__(*args, **kwargs))
    update = _immutable_error
    intersection_update = _immutable_error
    __ior__ = _immutable_error
    __iand__ = _immutable_error
    __ixor__ = _immutable_error
    difference_update = _immutable_error
    symmetric_difference_update = _immutable_error
    add = _immutable_error
    remove = _immutable_error
    discard = _immutable_error
    pop = _immutable_error
    clear = _immutable_error
    def __reduce__(self):
        return (
            hashableset,
            (set(self),)
        )




import inspect
import functools

def cached_property(fn):
    '''
    Used to decorate a function that should be a @property
    but also be cached. Worth careful consideration about
    whether we want to use this once we're fully on to Python 3.8.

    Main reason to use in place of functools.cached_property is
    this preserves the semantics of @property, as noted by the docs:
    > The mechanics of cached_property() are somewhat different from property().
    > A regular property blocks attribute writes unless a setter is defined.
    > In contrast, a cached_property allows writes.
    '''
    spec = inspect.getfullargspec(fn)
    assert spec.args == ['self']
    assert len(spec.kwonlyargs) == 0
    assert spec.varargs is None
    assert spec.varkw is None

    key = '_cached_'+fn.__name__
    @property
    @functools.wraps(fn)
    def wrapped(self):
        if not hasattr(self, key):
            setattr(self, key, fn(self))
        return getattr(self, key)
    return wrapped

def method_cache(fn):
    '''
    Avoiding functools.lru_cache since it holds onto object references
    when applied to methods of an object, creating a referenc cycle (not great for GC).
    Instead, preferring a strategy where cache storage lives on the object.
    '''
    cache_attr = '_cache_'+fn.__name__
    cache_info_attr = '_cache_info_'+fn.__name__
    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        # Since the store is object-local, we always have to ensure
        # objects passed in have the cache initialized.
        if not hasattr(self, cache_attr):
            setattr(self, cache_attr, {})
            setattr(self, cache_info_attr, dict(hits=0, misses=0))
        cache = getattr(self, cache_attr)
        cache_info = getattr(self, cache_info_attr)

        # Now we check for this function call, and
        # run the function if it hasn't been called before.
        key = (args, frozenset(kwargs.items()) if kwargs else None)
        if key not in cache:
            cache[key] = fn(self, *args, **kwargs)
            cache_info['misses'] += 1
        cache_info['hits'] += 1
        return cache[key]
    return wrapped



import sys
from typing import Tuple, Sequence, Union, Any, Callable, Generic, TypeVar
from dataclasses import is_dataclass, asdict
from itertools import combinations_with_replacement
from collections import Counter, defaultdict
import math

from flippy.tools import isclose, ISCLOSE_ATOL, ISCLOSE_RTOL, PackagePlaceholder
from functools import cached_property
from flippy.distributions.base import Distribution, FiniteDistribution, Element
from flippy.distributions.support import \
    Interval, Range, Simplex, OrderedIntegerPartitions, UnionSet
from flippy.distributions.random import default_rng

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = PackagePlaceholder("matplotlib.pyplot")

try:
    import numpy as np
except ImportError:
    np = PackagePlaceholder("numpy")

def beta_function(*alphas):
    num = math.prod(math.gamma(a) if not isclose(a, 0) else float('inf') for a in alphas)
    den = math.gamma(sum(alphas))
    return num/den


class Bernoulli(FiniteDistribution):
    '''
    A Bernoulli distribution. The support is a boolean value, with a single
    parameter `p` that specifies the probability of the value `True`.
    '''
    support = (True, False)
    def __init__(self, p=.5):
        self.p = p
    def sample(self, rng=default_rng, name=None, initial_value=None) -> bool:
        if rng.random() <= self.p:
            return True
        return False
    def log_probability(self, element):
        return {
            True: math.log(self.p) if self.p != 0.0 else float('-inf'),
            False: math.log(1 - self.p) if self.p != 1.0 else float('-inf')
        }.get(element, float('-inf'))
    def __repr__(self):
        return f"Bernoulli(p={self.p})"


MarginalElement = TypeVar('MarginalElement')

def is_namedtuple(obj):
    return (
        isinstance(obj, tuple) and
        hasattr(obj, '_asdict') and
        hasattr(obj, '_fields')
    )

class Categorical(FiniteDistribution[Element]):
    '''
    A distribution over a discrete support. Defaults to uniform probability of
    each item, unless `probabilities` or `weights` are specified.

    - `support` lists the discrete support.
    - `probabilities` is optional. Specifies the probabilities of each item of
    the support.
    - `weights` is optional. Specifies the weight for each item of the support.
    Transformed to probabilities by normalizing by the sum of weights.
    '''
    def __init__(self, support, *, probabilities=None, weights=None):
        if probabilities is not None:
            assert isclose(sum(probabilities), 1, atol=ISCLOSE_ATOL, rtol=ISCLOSE_RTOL)
        elif weights is not None:
            total_weight = sum(weights)
            probabilities = [w / total_weight for w in weights]
            del weights
        else:
            probabilities = [1/len(support) for _ in support]
        assert len(probabilities) == len(support), f'Expected the size of support ({len(support)}) to match the length of weights/probabilities ({len(probabilities)}), but they do not.'
        self.support = support
        self._probabilities = probabilities

    @classmethod
    def from_dict(cls, d: dict[Element, float]):
        support, probs = zip(*d.items())
        return cls(support=support, probabilities=probs)

    @classmethod
    def from_continuous(cls, dist: Distribution[Element], support: Sequence[float]):
        probs = [dist.prob(s) for s in support]
        total = sum(probs)
        probs = [p/total for p in probs]
        return cls(support=support, probabilities=probs)

    @property
    def probabilities(self):
        return self._probabilities

    def sample(self, rng=default_rng, name=None, initial_value=None) -> Element:
        return rng.choices(self.support, weights=self._probabilities, k=1)[0]

    def log_probability(self, element) -> float:
        try:
            return math.log(self._probabilities[self.support.index(element)])
        except ValueError:
            return float('-inf')

    def __repr__(self):
        return f"{self.__class__.__name__}(support={self.support}, probabilities={self._probabilities})"

    def _default_repr_html_(self):
        format_prob = lambda p: f"{p:.3f}" if p > 0.001 else f"{p:.2e}"
        return ''.join([
            "<table>",
            "<thead><tr><th></th><th>Element</th><th>Probability</th></tr></thead>",
            "<tbody>",
            *(
                [
                    f"<tr><td><b>{i}</b></td><td>{s}</td><td>{format_prob(self.prob(s))}</td></tr>"
                    for i, s in enumerate(self.support)
                ] if len(self.support) < 10 else (
                    [
                        f"<tr><td><b>{i}</b></td><td>{s}</td><td>{format_prob(self.prob(s))}</td></tr>"
                        for i, s in enumerate(self.support[:5])
                    ] + [
                        "<tr><td>...</td><td>...</td><td>...</td></tr>"
                    ] + [
                        f"<tr><td><b>{i}</b></td><td>{s}</td><td>{format_prob(self.prob(s))}</td></tr>"
                        for i, s in zip(range(len(self.support), len(self.support) - 5, -1),self.support[-5:])
                    ]
                )
            ),
            "</tbody>",
            "</table>"
        ])

    def _ReturnDict_repr_html(self):
        format_prob = lambda p: f"{p:.3f}" if p > 0.001 else f"{p:.2e}"
        keys = set()
        support = []
        probs = []
        for e in self.support:
            prob = self.prob(e)
            probs.append(prob)
            if is_dataclass(e):
                e = asdict(e)
            elif is_namedtuple(e):
                e = e._asdict()
            assert isinstance(e, dict)
            support.append(e)

        for e in support:
            keys.update(e.keys())
        def make_row(e : dict, p : float):
            row = []
            for k in keys:
                row.append(f"<td>{str(e.get(k, ''))}</td>")
            row.append(f"<td>{format_prob(p)}</td>")
            return f"<tr>{''.join(row)}</tr>"

        rows = []
        if len(support) < 10:
            for e, p in zip(support, probs):
                rows.append(make_row(e, p))
        else:
            for e, p in zip(support[:5], probs[:5]):
                rows.append(make_row(e, p))
            rows.append("<tr><td>...</td></tr>")
            for e, p in zip(support[-5:], probs[-5:]):
                rows.append(make_row(e, p))

        return ''.join([
            "<table>",
            "<thead><tr>"+''.join([f"<th>{k}</th>" for k in keys])+"<th>Probability</th></tr></thead>",
            "<tbody>",
            *rows,
            "</tbody>",
            "</table>"
        ])

    def _repr_html_(self):
        if (
            all(isinstance(e, dict) for e in self.support) or \
            all(is_dataclass(e) for e in self.support) or \
            all(is_namedtuple(e) for e in self.support)
        ):
            return self._ReturnDict_repr_html()
        return self._default_repr_html_()

    def plot(self, ax=None, bins=100, **kwargs):
        element = next(iter(self.support))
        if isinstance(element, (int, float)):
            return self._plot_1d(ax=ax, bins=bins, **kwargs)
        elif isinstance(element, (list, tuple)) and len(element) == 2:
            return self._plot_2d(ax=ax, bins=bins, **kwargs)
        elif isinstance(element, dict):
            if len(element) == 1:
                key = next(iter(element.keys()))
                dist = self.marginalize(lambda d : d[key])
                ax = dist._plot_1d(ax=ax, bins=bins, **kwargs)
                ax.set_xlabel(key)
                return ax
            if len(element) == 2:
                xkey, ykey = list(element.keys())
                dist = self.marginalize(lambda d : (d[xkey], d[ykey]))
                ax = dist._plot_2d(ax=ax, bins=bins, **kwargs)
                ax.set_xlabel(ykey)
                ax.set_ylabel(xkey)
                return ax
        elif is_dataclass(element):
            if len(element.__annotations__) == 1:
                key = next(iter(element.__annotations__))
                dist = self.marginalize(lambda d : getattr(d, key))
                ax = dist._plot_1d(ax=ax, bins=bins, **kwargs)
                ax.set_xlabel(key)
                return ax
            if len(element.__annotations__) == 2:
                xkey, ykey = element.__annotations__
                dist = self.marginalize(lambda d : (getattr(d, xkey), getattr(d, ykey)))
                ax = dist._plot_2d(ax=ax, bins=bins, **kwargs)
                ax.set_xlabel(ykey)
                ax.set_ylabel(xkey)
                return ax
        raise NotImplementedError("Can't plot this distribution")

    def _plot_2d(self, ax=None, bins=100, **kwargs):
        assert all(isinstance(s, (list, tuple)) and len(s) == 2 for s in self.support)
        if ax is None:
            fig, ax = plt.subplots()
        kwargs = {
            **{'origin': 'lower', 'aspect': 'auto'},
            **kwargs
        }
        xs, ys = zip(*self.support)
        hist, xedges, yedges = np.histogram2d(x=xs, y=ys, bins=bins, weights=self.probabilities)
        xlen = xedges[-1] - xedges[0]
        xmid = xlen/2
        xedges = [xmid - (xlen/2)*1.05, xmid + (xlen/2)*1.05]
        ymid = (yedges[0] + yedges[-1])/2
        ylen = yedges[-1] - yedges[0]
        yedges = [ymid - (ylen/2)*1.05, ymid + (ylen/2)*1.05]
        extent = [yedges[0], yedges[-1], xedges[0], xedges[-1]]
        ax.imshow(hist, extent=extent, **kwargs)
        cbar = ax.figure.colorbar(ax.images[0], ax=ax, location='right')
        cbar.set_label('Probability')
        return ax

    def _plot_1d(self, ax=None, bins=100, **kwargs):
        assert all(isinstance(s, (int, float)) for s in self.support)
        if ax is None:
            fig, ax = plt.subplots()
        ax.hist(self.support, weights=self.probabilities, bins=bins, **kwargs)
        ax.set_ylabel("Probability")
        return ax

    def marginalize(self, projection: Callable[[Element], MarginalElement]) -> 'Categorical[MarginalElement]':
        d = defaultdict(float)
        for s, p in zip(self.support, self.probabilities):
            d[projection(s)] += p
        return Categorical.from_dict(d)

    def condition(self, likelihood: Callable[[Element], float]) -> 'Categorical[Element]':
        posterior = {}
        for e, prior in self.items():
            l = likelihood(e)
            if l == 0:
                continue
            posterior[e] = prior * l
        norm = sum(posterior.values())
        return Categorical.from_dict({e: p/norm for e, p in posterior.items()})


class Multinomial(FiniteDistribution):
    '''
    This distribution takes multiple samples from a `Categorical` distribution.
    It takes the same parameters as `Categorical`, with the addition of `trials`,
    which are the number of samples drawn from the `Categorical`.
    '''
    def __init__(self, categorical_support, trials, *, probabilities=None, weights=None):
        self.categorical = Categorical(
            categorical_support,
            probabilities=probabilities,
            weights=weights
        )
        self.trials = trials

    def sample(self, rng=default_rng, name=None, initial_value=None) -> Tuple[int, ...]:
        samples = rng.choices(
            self.categorical.support,
            weights=self.categorical._probabilities,
            k=self.trials
        )
        counts = Counter(samples)
        return tuple(counts.get(i, 0) for i in self.categorical.support)

    @cached_property
    def support(self):
        return OrderedIntegerPartitions(
            total=self.trials, partitions=len(self.categorical.support)
        )

    def log_probability(self, vec):
        if vec in self.support:
            probs = self.categorical._probabilities
            num1 = math.gamma(self.trials + 1)
            num2 = math.prod(p**x for p, x in zip(probs, vec))
            den = math.prod(math.gamma(x + 1) for x in vec)
            return math.log((num1*num2)/den)
        return float('-inf')


class Gaussian(Distribution):
    r'''
    A Gaussian distribution, with support over the real numbers $\mathbb{R}$.

    - `mean` is the mean.
    - `sd` is the standard deviation.
    '''

    support = Interval(float('-inf'), float('inf'), left_open=True, right_open=True)
    def __init__(self, mean=0, sd=1):
        self.mean = mean
        self.sd = sd

    def sample(self, rng=default_rng, name=None, initial_value=None) -> float:
        return rng.gauss(self.mean, self.sd)

    def log_probability(self, element):
        prob = (
            math.e**(-.5*((element - self.mean)/self.sd)**2)
        )/(
            self.sd*(2*math.pi)**.5
        )
        return math.log(prob) if prob > 0. else float('-inf')
Normal = Gaussian


class Uniform(Distribution):
    '''
    A Uniform distribution, with support over a specified interval [`start`, `end`].

    - `start` defaults to 0. The lower bound of the interval.
    - `end` defaults to 1. The upper bound of the interval.
    '''

    def __init__(self, start=0, end=1):
        self.start = start
        self.end = end
        self.support = Interval(start, end, left_open=False, right_open=False)

    def sample(self, rng=default_rng, name=None, initial_value=None) -> float:
        return rng.uniform(self.start, self.end)

    def log_probability(self, element):
        if element in self.support:
            return math.log(1/(self.end - self.start))
        return float('-inf')


class Beta(Distribution):
    """
    A Beta distribution, with support over the interval [0, 1].
    - `alpha` is the first shape parameter (default 1).
    - `beta` is the second shape parameter (default 1).
    """
    support = Interval(0, 1, left_open=False, right_open=False)
    def __init__(self, alpha=1, beta=1):
        self.a = alpha
        self.b = beta

    def sample(self, rng=default_rng, name=None, initial_value=None) -> float:
        return rng.betavariate(self.a, self.b)

    def log_probability(self, element):
        if element in self.support:
            # to avoid numerical issues at the boundaries
            element = max(min(element, 1 - sys.float_info.epsilon), sys.float_info.epsilon)
            num = (element**(self.a - 1))*(1 - element)**(self.b - 1)
            prob = num/beta_function(self.a, self.b)
            return math.log(prob) if prob != 0 else float('-inf')
        return float('-inf')

    def plot(self, ax=None, bins=100, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()
        x = np.linspace(0, 1, 1000)
        ax.plot(x, [self.prob(i) for i in x], **kwargs)
        return ax


class Binomial(Distribution):
    """
    A Binomial distribution, with support over the integers [0, trials].

    - `trials` is the number of trials
    - `p` is the probability of success on each trial
    """

    def __init__(self, trials : int, p : float):
        self.trials = trials
        self.p = p
        assert 0 <= p <= 1
        self.support = tuple(range(0, self.trials + 1))

    def sample(self, rng=default_rng, name=None, initial_value=None) -> int:
        return sum(rng.random() < self.p for _ in range(self.trials))

    def log_probability(self, element):
        if element in self.support:
            prob = (
                math.comb(self.trials, element)
            )*(
                self.p**element
            )*(
                (1 - self.p)**(self.trials - element)
            )
            return math.log(prob) if prob != 0 else float('-inf')
        return float('-inf')


class Geometric(Distribution):
    """
    A Geometric distribution, with support over the non-negative integers {0, 1, 2, ...}.
    - `p` is the probability of success on each trial
    """

    support = Range(0, float('inf'))
    def __init__(self, p : float):
        self.p = p
        assert 0 <= p <= 1

    def sample(self, rng=default_rng, name=None, initial_value=None) -> int:
        i = 0
        while rng.random() >= self.p:
            i += 1
        return i

    def log_probability(self, element):
        if element in self.support:
            return math.log(
                ((1 - self.p)**(element))*self.p
            )
        return float('-inf')


class Poisson(Distribution):
    """
    A Poisson distribution, with support over the non-negative integers {0, 1, 2, ...}.
    """

    support = Range(0, float('inf'))
    def __init__(self, rate : float):
        self.rate = rate
        assert rate >= 0

    def sample(self, rng=default_rng, name=None, initial_value=None) -> int:
        p, k, L = 1, 0, math.exp(-self.rate)
        while p > L:
            k += 1
            p = rng.random()*p
        return k - 1

    def log_probability(self, k):
        if k not in self.support:
            return float('-inf')
        prob = (self.rate**k)*math.exp(-self.rate)/math.factorial(k)
        return math.log(prob)

class Gamma(Distribution):
    """
    A Gamma distribution, with support over the positive real numbers (0, inf).
    """

    support = Interval(0, float('inf'), left_open=False, right_open=True)
    def __init__(self, shape, rate):
        self.shape = shape
        self.rate = rate

    def sample(self, rng=default_rng, name=None, initial_value=None) -> float:
        # uses the shape, scale parameterization
        return rng.gammavariate(self.shape, 1/self.rate)

    def log_probability(self, element):
        if element in self.support:
            prob = (
                (self.rate**self.shape)*\
                (element**(self.shape - 1))*\
                math.exp(-self.rate*element)
            )/(
                math.gamma(self.shape)
            )
            return math.log(prob) if prob != 0 else float('-inf')
        return float('-inf')

class BetaBinomial(FiniteDistribution):
    """
    A Beta-Binomial compound distribution, with support over the integers [0, trials].
    """

    def __init__(self, trials : int, alpha=1, beta=1):
        self.alpha = alpha
        self.beta = beta
        self.trials = trials
        self.support = tuple(range(0, self.trials + 1))

    def sample(self, rng=default_rng, name=None, initial_value=None) -> int:
        p = rng.betavariate(self.alpha, self.beta)
        return sum(rng.random() < p for _ in range(self.trials))

    def log_probability(self, element):
        if element in self.support:
            prob = math.comb(self.trials, element)*(
                (
                    beta_function(element+self.alpha, self.trials - element + self.beta)
                )/(
                    beta_function(self.alpha, self.beta)
                )
            )
            return math.log(prob)
        return float('-inf')

    def update(self, data: Sequence[int]) -> 'BetaBinomial':
        assert all(0 <= d <= self.trials for d in data)
        new_alpha = self.alpha + sum(data)
        new_beta = self.beta + self.trials*len(data) - sum(data)
        return BetaBinomial(trials=self.trials, alpha=new_alpha, beta=new_beta)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(trials={self.trials}, alpha={self.alpha}, beta={self.beta})"


class Dirichlet(Distribution):
    """
    A Dirichlet distribution, with support over the (k-1)-simplex.
    - `alphas` is a sequence of shape parameters, one for each dimension of the simplex.
    """

    def __init__(self, alphas):
        self.alphas = alphas

    @cached_property
    def support(self):
        return Simplex(len(self.alphas))

    def sample(self, rng=default_rng, name=None, initial_value=None) -> Tuple[float, ...]:
        e = [rng.gammavariate(a, 1) for a in self.alphas]
        e = [ei if ei > 0 else sys.float_info.min for ei in e] # for numerical stability
        tot = sum(e)
        vals = tuple(ei/tot for ei in e)
        if 1 in vals:
            return tuple(min(ei, 1.0 - sys.float_info.epsilon) for ei in vals)
        return vals

    def log_probability(self, vec):
        r"""
        Log probability of a vector in the Dirichlet distribution:
        $$
        \ln P(x) = \sum_{i=1}^{k} (a_i - 1) \ln x_i - \ln B(a_1, a_2, \ldots, a_k)
        $$
        """
        if vec in self.support:
            log_num = sum((a - 1) * math.log(v) for v, a in zip(vec, self.alphas))
            log_den = math.log(beta_function(*self.alphas))
            return log_num - log_den
        return float('-inf')

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(alphas={self.alphas})'


class DirichletMultinomial(FiniteDistribution):
    """
    A Dirichlet-Multinomial compound distribution, with support over the
    integer partitions of `trials` into `len(alphas)` parts.
    - `trials` is the number of trials
    - `alphas` is a sequence of shape parameters, one for each category.
    """

    def __init__(self, trials : int, alphas : Sequence[float]):
        self.alphas = alphas
        self.trials = trials

    @cached_property
    def support(self) -> OrderedIntegerPartitions:
        return OrderedIntegerPartitions(
            total=self.trials, partitions=len(self.alphas)
        )

    @cached_property
    def dirichlet(self) -> Dirichlet:
        return Dirichlet(self.alphas)

    def sample(self, rng=default_rng, name=None, initial_value=None) -> Tuple[int, ...]:
        ps = self.dirichlet.sample(rng=rng)
        samples = rng.choices(range(len(self.alphas)), weights=ps, k=self.trials)
        counts = Counter(samples)
        return tuple(counts.get(i, 0) for i in range(len(self.alphas)))

    def log_probability(self, vec : Tuple[int, ...]) -> float:
        if vec not in self.support:
            return float('-inf')
        assert len(vec) == len(self.alphas)
        num = self.trials*beta_function(sum(self.alphas), self.trials)
        den = math.prod(
            x*beta_function(a, x) for a, x in zip(self.alphas, vec)
            if x > 0
        )
        return math.log(num/den)

    def update(self, data : Sequence[Tuple[int, ...]]) -> 'DirichletMultinomial':
        assert all(len(d) == len(self.alphas) for d in data)
        assert all(sum(d) == self.trials for d in data)
        cat_counts = zip(*data)
        new_alphas = tuple(a + sum(c) for a, c in zip(self.alphas, cat_counts))
        return DirichletMultinomial(trials=self.trials, alphas=new_alphas)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(trials={self.trials}, alphas={self.alphas})"

class Mixture(Distribution):
    """
    A mixture distribution, which is a weighted combination of multiple distributions.
    - `distributions` is a sequence of `Distribution` instances to mix.
    - `weights` is an optional sequence of weights for each distribution. If not provided,
      uniform weights are assumed. Weights must sum to 1.

    - Alternatively, use the class method `from_distribution_of_distributions`
    to create a mixture from a `FiniteDistribution` over distributions.
    """

    def __init__(
        self,
        distributions : Sequence[Distribution],
        weights : Sequence[float] = None
    ):
        if weights is None:
            weights = [1 / len(distributions) for _ in distributions]
        assert len(distributions) == len(weights), "Must have a weight for each distribution"
        assert isclose(sum(weights), 1), "Weights must sum to 1"
        self.distributions = distributions
        self.weights = weights

    @classmethod
    def from_distribution_of_distributions(
        cls,
        dist_of_dists : FiniteDistribution[Distribution]
    ):
        probabilities = [dist_of_dists.prob(d) for d in dist_of_dists.support]
        return cls(
            distributions=dist_of_dists.support,
            weights=probabilities
        )

    @cached_property
    def support(self):
        return UnionSet(*(d.support for d in self.distributions))

    def sample(self, rng=default_rng, name=None, initial_value=None) -> Element:
        dist = rng.choices(self.distributions, weights=self.weights)[0]
        return dist.sample(rng=rng, name=name)

    def log_probability(self, element : Element) -> float:
        total_prob = 0
        for dist, weight in zip(self.distributions, self.weights):
            total_prob += weight * dist.prob(element)
        if total_prob == 0:
            return float('-inf')
        return math.log(total_prob)

    def expected_value(self, func: Callable[[Element], float] = lambda v : v) -> float:
        total_expected_value = 0
        for dist, weight in zip(self.distributions, self.weights):
            total_expected_value += weight * dist.expected_value(func=func)
        return total_expected_value

    def plot(self, *, ax=None, xmin=None, xmax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()
        if xmin is None:
            xmin = min([d.support.start for d in self.distributions])
        if xmax is None:
            xmax = max([d.support.end for d in self.distributions])
        x = np.linspace(xmin, xmax, 1000)
        ax.plot(x, [self.prob(i) for i in x], **kwargs)
        return ax



from collections.abc import Iterable
from typing import Sequence, Generic, TypeVar, Any, Callable
import math
import random
import abc
from functools import cached_property

from flippy.distributions.support import Support
from flippy.types import Element
from flippy.hashable import hashabledict
from flippy.tools import isclose, ISCLOSE_RTOL, ISCLOSE_ATOL
from flippy.transforms import CPSTransform

class Distribution(Generic[Element]):
    """
    Base class for probability distributions in FlipPy.
    """
    support : Support
    def prob(self, element : Element):
        """Probability of an element in the support of the distribution."""
        return math.exp(self.log_probability(element))

    @abc.abstractmethod
    def sample(self, rng=random, name=None, initial_value=None) -> Element:
        """Sample an element from the distribution."""
        pass

    def observe(self, value: Element) -> None:
        """Observe a value, which is equivalent to conditioning on that value."""
        pass

    def fit(self, *, rng=random, name=None, initial_value: Element = None):
        """Variables returned by this method are fit to maximize marginal likelihood."""
        return initial_value

    @abc.abstractmethod
    def log_probability(self, element : Element) -> float:
        """Log probability of an element"""
        pass

    def expected_value(self, func: Callable[[Element], float] = lambda v : v) -> float:
        """Expected value of the distribution, optionally transformed by `func`."""
        raise NotImplementedError

    def isclose(self, other: "Distribution", *, rtol: float=ISCLOSE_RTOL, atol: float=ISCLOSE_ATOL) -> bool:
        """Check if two distributions are close in terms of their log probabilities."""
        raise NotImplementedError

    def plot(self, ax=None, **kws):
        raise NotImplementedError

    def update(self, data : Sequence[Element]):
        raise NotImplementedError

    def __bool__(self):
        raise ValueError("Cannot convert distribution to bool")

    def total_log_probability(self, data : Sequence[Element]) -> float:
        """Calculate the total log probability of a sequence of data points."""
        return sum(self.log_probability(d) for d in data)
    setattr(total_log_probability, CPSTransform.is_transformed_property, True)

    # This method will be CPS transformed
    def observe_all(self, data : Iterable[Element]):
        """
        Observe all elements in a sequence, equivalent to conditioning on each element.
        """
        _factor_dist.observe(self.total_log_probability(data))


class FactorDistribution(Distribution):
    def __init__(self):
        pass

    def sample(self, rng, name, initial_value=None):
        return 0

    def log_probability(self, element : float) -> float:
        #workaround for arbitrary scores
        return element

_factor_dist = FactorDistribution()


class FiniteDistribution(Distribution[Element]):
    support: Sequence[Element]

    @cached_property
    def probabilities(self):
        return tuple(self.prob(e) for e in self.support)

    def isclose(self, other: "FiniteDistribution", *, rtol: float=ISCLOSE_RTOL, atol: float=ISCLOSE_ATOL) -> bool:
        full_support = set(self.support) | set(other.support)
        return all(
            isclose(self.log_probability(s), other.log_probability(s), rtol=rtol, atol=atol)
            for s in full_support
        )

    def items(self):
        yield from zip(self.support, self.probabilities)

    def expected_value(self, func: Callable[[Element], Any] = lambda v : v) -> Any:
        return sum(
            p*func(s)
            for s, p in self.items()
        )

    def __getitem__(self, element):
        return self.prob(element)

    def as_dict(self):
        return dict(zip(self.support, self.probabilities))

    def __len__(self):
        return len(self.support)

    def keys(self):
        yield from self.support

    def values(self):
        yield from self.probabilities

    def items(self):
        yield from zip(self.support, self.probabilities)

    def __iter__(self):
        yield from self.support

    def __hash__(self):
        return hash(hashabledict(self.as_dict()))

    def __eq__(self, other):
        if not isinstance(other, FiniteDistribution):
            return False
        return self.as_dict() == other.as_dict()



import math
import random
from itertools import product
from flippy.distributions.builtin_dists import Categorical, Bernoulli, \
    Multinomial, DirichletMultinomial, Dirichlet, Multinomial, Beta, \
    Binomial, BetaBinomial
from flippy.inference import LikelihoodWeighting
from flippy.distributions.random import RandomNumberGenerator
from flippy.tools import isclose

def test_distribution_isclose():
    # Same distribution.
    a = Categorical(range(3))
    b = Categorical(range(3))
    assert a.isclose(b)
    assert b.isclose(a)

    # Mismatched probabilities.
    a = Categorical(range(3))
    b = Categorical(range(3), probabilities=[.1, .4, .5])
    assert not a.isclose(b)
    assert not b.isclose(a)

    # Mismatched support, but probabilities match on non-zero support.
    a = Categorical(range(2))
    b = Categorical(range(3), probabilities=[.5, .5, 0])
    assert a.isclose(b)
    assert b.isclose(a)

    # Mismatched support.
    a = Categorical(range(2))
    b = Categorical(range(3))
    assert not a.isclose(b)
    assert not b.isclose(a)

    # Different distributions, but match.
    a = Categorical([True, False])
    b = Bernoulli(0.5)
    assert a.isclose(b)
    assert b.isclose(a)

    # Matched with an absolute tolerance
    a = Bernoulli(0.5 + 1e-10)
    b = Bernoulli(0.5)
    assert a.isclose(b)

    # Tolerance can be changed
    a = Bernoulli(0.50001)
    b = Bernoulli(0.5)
    assert not a.isclose(b)
    assert not a.isclose(b, atol=1e-5)
    assert a.isclose(b, atol=1e-4)
    assert a.isclose(b, atol=1e-3)

def test_Multinomial_pdf():
    balls_bins = [
        (5, 6), (1, 5), (3, 3)
    ]
    for balls, bins in balls_bins:
        support = list(product(range(balls + 1), repeat=bins))
        dist = Multinomial(categorical_support=range(bins), trials=balls)
        tot = sum([math.exp(dist.log_probability(v)) for v in support])
        assert isclose(1.0, tot)

def test_DirichletMultinomial_pdf():
    balls_bins = [
        (5, 6), (1, 5), (3, 3)
    ]
    for balls, bins in balls_bins:
        support = list(product(range(balls + 1), repeat=bins))
        dist = DirichletMultinomial(
                trials=balls,
                alphas=[random.random()*1 for _ in range(bins)]
            )
        prob = lambda v : math.exp(dist.log_probability(v))
        tot = sum([prob(v) for v in support])
        assert isclose(1.0, tot), (tot, (balls, bins))

def test_dirichletmultinomial_update():
    data = [(0, 1, 0), (0, 1, 0), (0, 0, 1), (0, 0, 1)]

    def model():
        p = Dirichlet([1, 1, 1]).sample()
        dist = Multinomial(categorical_support=range(len(p)), trials=1, probabilities=p)
        [dist.observe(d) for d in data]
        return dist.sample()

    seed = 139491
    res = LikelihoodWeighting(model, samples=2000, seed=seed).run()

    dist = DirichletMultinomial(trials=1, alphas=[1, 1, 1])
    dist2 = dist.update(data)

    for e in dist2.support:
        assert isclose(dist2.prob(e), res.prob(e), atol=.01)

def test_betabinomial_update():
    data = [0]*3 + [1]*2 + [2]*1

    def model():
        p = Beta(1, 2).sample()
        dist = Binomial(trials=3, p=p)
        [dist.observe(d) for d in data]
        return dist.sample()

    seed = 139421
    res = LikelihoodWeighting(model, samples=10000, seed=seed).run()

    dist = BetaBinomial(trials=3, alpha=1, beta=1)
    dist2 = dist.update(data)
    for e in dist2.support:
        assert isclose(dist2.prob(e), res.prob(e), atol=.05)


def test_random_number_generation():
    rng1a = RandomNumberGenerator(20)
    rng1b = RandomNumberGenerator(20)
    rng1c = RandomNumberGenerator(20)
    assert rng1a.randint(0, 100) == rng1b.randint(0, 100)
    assert rng1a != rng1b != rng1c
    assert rng1a.randint(0, 100) != rng1c.randint(0, 100)
    rng2 = RandomNumberGenerator(30)
    rng3 = RandomNumberGenerator(40)
    assert rng2.randint(0, 100) != rng3.randint(0, 100)

def test_random_number_numpy():
    rng1 = RandomNumberGenerator(30)
    rng2 = RandomNumberGenerator(30)
    rng3 = RandomNumberGenerator(40)
    assert rng1.np.random() == rng2.np.random()
    assert rng1.np.random() != rng3.np.random()



from typing import Sequence, Tuple, Callable, TypeVar, List, TYPE_CHECKING

from flippy.core import ProgramState
from flippy.types import Continuation, CPSCallable
from flippy.interpreter import CPSInterpreter
from flippy.transforms import CPSTransform

from flippy.callentryexit import register_call_entryexit
if TYPE_CHECKING:
    from flippy.interpreter import Stack

T = TypeVar('T')
I = TypeVar('I')

def recursive_map(fn: Callable[[I],T], iter: Sequence[I]) -> List[T]:
    if not iter:
        return []
    return [fn(iter[0])] + recursive_map(fn, iter[1:])

def independent_map(func: Callable[[I],T], iterator: Sequence[I]) -> Tuple[T, ...]:
    """
    This gives the inference algorithm access to a mapped function
    applied to each iterate using the EnterCallState/ExitCallState interface.
    We do this by:

    1. Sending a `MapEnter` signal to the algorithm to see if it wants access to
    the mapped function applied to each iterate. If not, it just uses `recursive_map`.

    2. If it does, we intercept and then reinstate the CPS transform to construct a
    sequence of EnterCallState/ExitCallState blocks that
    *does not save non-global state between each iterate*. The inference algorithm
    is responsible for handing return values from `ExitCallState`s.

    3. Sending a `MapExit` signal to the algorithm that accepts the result of applying
    the mapped function over each iterate.
    """
    send_call_entryexit = map_enter_event()
    if send_call_entryexit:
        return _independent_map(func, iterator)
    else:
        return tuple(recursive_map(func, iterator))

class MapEnter(ProgramState):
    pass

class MapExit(ProgramState):
    pass

def map_enter_event(*, _stack=None, _cps=None, _cont=None):
    return MapEnter(
        continuation=lambda send_call_entryexit=False : _cont(send_call_entryexit),
        stack=_stack,
        cps=_cps,
    )
setattr(map_enter_event, CPSTransform.is_transformed_property, True)

def _independent_map(
    func: CPSCallable,
    iterator: Sequence,
    *,
    _stack: 'Stack' = None,
    _cps: CPSInterpreter = None,
    _cont: Continuation = None
):
    def construct_cont(next_cont, i):
        # note we don't actually use the return value of each iterate
        # here (_res). The inference algorithm is responsible for handling
        # that when it receives an `ExitCallState` message.
        return lambda : _cps.interpret(
            _independent_map_iter,
            cont=lambda _res : next_cont(),
            stack=_stack,
            func_src="INDEPENDENT_MAP_ITER",
            locals_={},
            lineno=0
        )(func, i)

    # we start from the end and then iterate backwards
    next_cont = lambda : MapExit(
        continuation=lambda map_result: _cont(map_result),
        stack=_stack,
        cps=_cps,
    )
    for i in iterator[::-1]:
        last_cont = construct_cont(next_cont, i)
        next_cont = last_cont
    return last_cont
setattr(_independent_map, CPSTransform.is_transformed_property, True)

# Here's where each individual function call lives. It is wrapped in an entry/exit
# wrapper and gets CPS transformed.
def _independent_map_iter(func, i):
    return func(i)
_independent_map_iter = CPSInterpreter().non_cps_callable_to_cps_callable(_independent_map_iter)
_independent_map_iter = register_call_entryexit(_independent_map_iter)
